# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

project = 'mols2grid'
copyright = '2022, Cédric Bouysset'
author = 'Cédric Bouysset'


# -- General configuration ---------------------------------------------------

github_doc_root = 'https://github.com/cbouy/mols2grid/tree/master/docs/'
needs_sphinx = '4.5.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode', 'sphinx.ext.napoleon',
    'sphinx.ext.autosectionlabel',
    'sphinx_rtd_theme',
    'sphinx_mdinclude',
    'nbsphinx',
]

autosectionlabel_prefix_document = True
napoleon_google_docstring = False

source_suffix = ['.rst', '.md']

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
pygments_style = "manni"
html_logo = "_static/mols2grid_logo.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

###

intersphinx_mapping = {'https://docs.python.org/3/': None,
                       'https://numpy.org/doc/stable/': None,
                       'https://www.rdkit.org/docs/': None,
                       'https://pandas.pydata.org/docs/': None,
                       'https://ipython.readthedocs.io/en/stable/': None,
}

# app setup hook
def setup(app):
    # custom css
    app.add_css_file('custom.css')



from ipywidgets import DOMWidget, register
from traitlets import Bool, List, Unicode

from ._frontend import module_name, module_version


@register
class MolGridWidget(DOMWidget):
    """A custom widget for the MolGrid class. Handles selections and callbacks.

    Attributes
    ----------
    grid_id : str
        Name of the grid controlling the widget
    selection : str
        JSON string containing the molecule selection as a dictionnary. Index
        are keys and SMILES string are values.
    callback_kwargs : str
        JSON string containing the keyword arguments with which to call the
        callback function.
    filter_mask : List[bool]
        List stating wether a molecule should be kept (True) or filtered out
        (False)
    """

    _model_name = Unicode("MolGridModel").tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode("MolGridView").tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    grid_id = Unicode("default").tag(sync=True)
    selection = Unicode("{}").tag(sync=True)
    callback_kwargs = Unicode("{}").tag(sync=True)
    filter_mask = List(Bool(), []).tag(sync=True)



from pathlib import Path

import pytest
from rdkit import RDConfig

from mols2grid import MolGrid, sdf_to_dataframe


@pytest.fixture(scope="module")
def sdf_path():
    return Path(RDConfig.RDDocsDir) / "Book" / "data" / "solubility.test.sdf"


@pytest.fixture(scope="module")
def sdf_file(sdf_path):
    return str(sdf_path)


@pytest.fixture(scope="module")
def smiles_records():
    return [{"SMILES": "C" * i, "ID": i} for i in range(1, 5)]


@pytest.fixture(scope="module")
def df(sdf_path):
    return sdf_to_dataframe(sdf_path).head(30)


@pytest.fixture(scope="module")
def small_df(df):
    return df.head(5)


@pytest.fixture(scope="module")
def grid_prerendered(df):
    return MolGrid(df, mol_col="mol", prerender=True)


@pytest.fixture(scope="module")
def grid(df):
    return MolGrid(df, mol_col="mol", size=(160, 120))


@pytest.fixture(scope="module")
def mols(small_df):
    return small_df["mol"]



import gzip
from tempfile import NamedTemporaryFile
from types import SimpleNamespace
from unittest.mock import Mock, patch

import pandas as pd
import pytest
from rdkit import Chem
from rdkit.Chem.rdDepictor import Compute2DCoords

from mols2grid import utils


def test_requires():
    @utils.requires("_not_a_module")
    def func():
        pass

    with pytest.raises(
        ModuleNotFoundError, match="The module '_not_a_module' is required"
    ):
        func()


@pytest.mark.parametrize(
    ["subset", "fmt", "style", "transform", "exp"],
    [
        (
            ["SMILES", "ID"],
            "<strong>{key}</strong>: {value}",
            {},
            {},
            '<strong>SMILES</strong>: <span class="copy-me">CCO</span><br>'
            '<strong>ID</strong>: <span class="copy-me">0</span>',
        ),
        (["ID"], "foo-{value}", {}, {}, 'foo-<span class="copy-me">0</span>'),
        (
            ["ID"],
            "{value}",
            {"ID": lambda x: "color: red"},
            {},
            '<span class="copy-me" style="color: red">0</span>',
        ),
        (
            ["Activity"],
            "{value}",
            {},
            {"Activity": lambda x: f"{x:.2f}"},
            '<span class="copy-me">42.01</span>',
        ),
        (
            ["Activity"],
            "{key}: {value}",
            {"Activity": lambda x: "color: red" if x > 40 else ""},
            {"Activity": lambda x: f"{x:.2f}"},
            'Activity: <span class="copy-me" style="color: red">42.01</span>',
        ),
    ],
)
def test_tooltip_formatter(subset, fmt, style, transform, exp):
    row = pd.Series(
        {
            "ID": 0,
            "SMILES": "CCO",
            "Activity": 42.012345,
        }
    )
    tooltip = utils.tooltip_formatter(row, subset, fmt, style, transform)
    assert tooltip == exp


@pytest.mark.parametrize(
    ["smi", "exp"], [("CCO", "CCO"), ("blabla", None), (None, None)]
)
def test_mol_to_smiles(smi, exp):
    if smi:
        mol = Chem.MolFromSmiles(smi)
    else:
        mol = smi
    assert utils.mol_to_smiles(mol) == exp


def test_mol_to_record():
    mol = Chem.MolFromSmiles("CCO")
    props = {
        "NAME": "ethanol",
        "foo": 42,
        "_bar": 42.01,
        "__baz": 0,
    }
    for prop, value in props.items():
        if isinstance(value, int):
            mol.SetIntProp(prop, value)
        elif isinstance(value, float):
            mol.SetDoubleProp(prop, value)
        else:
            mol.SetProp(prop, value)
    new = utils.mol_to_record(mol)
    assert "mol" in new.keys()
    new.pop("mol")
    assert new == props


def test_mol_to_record_none():
    new = utils.mol_to_record(None)
    assert new == {}


def test_mol_to_record_overwrite_smiles():
    mol = Chem.MolFromSmiles("CCO")
    mol.SetProp("SMILES", "foo")
    new = utils.mol_to_record(mol)
    assert new["SMILES"] == "foo"


def test_mol_to_record_custom_mol_col():
    mol = Chem.MolFromSmiles("CCO")
    new = utils.mol_to_record(mol, mol_col="foo")
    assert new["foo"] is mol


@pytest.mark.parametrize("sdf_source", ["sdf_path", "sdf_file"])
def test_sdf_to_dataframe(sdf_source, request):
    sdf = request.getfixturevalue(sdf_source)
    df = utils.sdf_to_dataframe(sdf)
    exp = {
        "ID": 5,
        "NAME": "3-methylpentane",
        "SMILES": "CCC(C)CC",
        "SOL": -3.68,
        "SOL_classification": "(A) low",
        "_MolFileComments": "",
        "_MolFileInfo": "  SciTegic05121109362D",
        "_Name": "3-methylpentane",
    }
    new = df.iloc[0].drop(["mol"]).to_dict()
    assert new == exp


def test_sdf_to_dataframe_custom_mol_col(sdf_path):
    df = utils.sdf_to_dataframe(sdf_path, mol_col="foo")
    assert "mol" not in df.columns
    assert "foo" in df.columns


def test_sdf_to_df_gz(sdf_path, tmp_path):
    tmp_file = tmp_path / "mols.sdf.gz"
    with open(sdf_path, "rb") as fi, open(tmp_file, "wb") as tf:
        gz = gzip.compress(fi.read(), compresslevel=1)
        tf.write(gz)
        tf.flush()
        df = utils.sdf_to_dataframe(tmp_file).drop(columns=["mol"])
        ref = utils.sdf_to_dataframe(sdf_path).drop(columns=["mol"])
        assert (df == ref).values.all()


def test_remove_coordinates():
    mol = Chem.MolFromSmiles("CCO")
    Compute2DCoords(mol)
    mol.GetConformer()
    new = utils.remove_coordinates(mol)
    assert new is mol
    with pytest.raises(ValueError, match="Bad Conformer Id"):
        new.GetConformer()


@pytest.mark.parametrize(
    ["string", "expected"],
    [
        ("Mol", "Mol"),
        ("mol name", "mol-name"),
        ("mol  name", "mol-name"),
        ("mol-name", "mol-name"),
        ("mol- name", "mol--name"),
        ("mol\tname", "mol-name"),
        ("mol\nname", "mol-name"),
        ("mol \t\n name", "mol-name"),
    ],
)
def test_slugify(string, expected):
    assert utils.slugify(string) == expected


@pytest.mark.parametrize("value", [1, 2])
def test_callback_handler(value):
    callback = lambda x: x + 1
    mock = Mock(side_effect=callback)
    event = SimpleNamespace(new=str(value))
    utils.callback_handler(mock, event)
    mock.assert_called_once_with(value)


def test_is_running_within_streamlit():
    assert utils.is_running_within_streamlit() is False
    with patch(
        "mols2grid.utils._get_streamlit_script_run_ctx",
        create=True,
        new=lambda: object(),
    ):
        assert utils.is_running_within_streamlit() is True
    with patch(
        "mols2grid.utils._get_streamlit_script_run_ctx", create=True, new=lambda: None
    ):
        assert utils.is_running_within_streamlit() is False



from types import SimpleNamespace

import pytest

import mols2grid as mg
from mols2grid.select import register


@pytest.fixture(autouse=True)
def clear_register_between_tests():
    register._clear()
    register._init_grid("foo")
    yield
    register._clear()


def test_clear_register():
    register._clear()
    assert not hasattr(register, "current_selection")
    assert register.SELECTIONS == {}


def test_update_current_grid(smiles_records):
    mg.MolGrid(smiles_records, name="bar")
    assert register.current_selection == "bar"


def test_init_grid():
    assert "foo" in register.SELECTIONS.keys()
    assert register.current_selection == "foo"


def test_overwrite_warning():
    event = SimpleNamespace(new='{0: "C"}')
    register.selection_updated("foo", event)
    with pytest.warns(
        UserWarning, match=f"Overwriting non-empty 'foo' grid selection: {{0: 'C'}}"
    ):
        register._init_grid("foo")
    assert register.get_selection() == {}


def test_update_and_get_selection():
    assert register.get_selection() == {}
    event = SimpleNamespace(new='{0: "CCO"}')
    register.selection_updated("foo", event)
    assert register.get_selection() == {0: "CCO"}
    event.new = "{}"
    register.selection_updated("foo", event)
    assert register.get_selection() == {}


def test_list_grids():
    assert register.list_grids() == ["foo"]
    register._init_grid("bar")
    assert register.list_grids() == ["foo", "bar"]
    register._init_grid("foo")
    assert register.list_grids() == ["foo", "bar"]



import json
import os
import sys
from base64 import b64encode
from datetime import datetime, timezone
from pathlib import Path
from types import SimpleNamespace

import imagehash
import pyautogecko
import pytest
from flaky import flaky
from rdkit import Chem
from rdkit import __version__ as rdkit_version
from rdkit.Chem import AllChem
from selenium import webdriver
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

import mols2grid
from mols2grid.select import register
from mols2grid.utils import env

from .webdriver_utils import FirefoxDriver

try:
    from rdkit.Chem.rdDepictor import IsCoordGenSupportAvailable
except ImportError:
    COORDGEN_SUPPORT = not sys.platform.startswith("win")
else:
    COORDGEN_SUPPORT = IsCoordGenSupportAvailable()

# for local debug, switch headless mode to False
HEADLESS = True

pytestmark = pytest.mark.webdriver
GITHUB_ACTIONS = bool(os.environ.get("GITHUB_ACTIONS"))
PAGE_LOAD_TIMEOUT = 10
skip_no_coordgen = pytest.mark.skipif(
    not COORDGEN_SUPPORT,
    reason="CoordGen library not available in this RDKit build",
)
pyautogecko.install()


def determine_scope(fixture_name, config):
    if GITHUB_ACTIONS:
        return "function"
    return "module"


def get_grid(df, **kwargs):
    kwargs.setdefault("mol_col", "mol")
    return mols2grid.MolGrid(df, **kwargs)


def get_doc(grid, kwargs):
    html = grid.render(**kwargs)
    html = b64encode(html.encode()).decode()
    return "data:text/html;base64,{}".format(html)


@pytest.fixture(scope=determine_scope)
def driver():
    options = webdriver.FirefoxOptions()
    options.headless = True if GITHUB_ACTIONS else HEADLESS
    driver = FirefoxDriver(options=options)
    driver.set_page_load_timeout(PAGE_LOAD_TIMEOUT)
    yield driver
    driver.quit()


@pytest.fixture(scope="module")
def html_doc(grid):
    return get_doc(
        grid,
        dict(
            n_items_per_page=5,
            subset=["_Name", "img"],
            size=(160, 120),
        ),
    )


# make sure non-parametrized test is ran first
@pytest.mark.order(1)
def test_no_subset_all_visible(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"tooltip": [], "selection": False})
    driver.get(doc)
    columns = set(grid.dataframe.columns.drop(["mol", "mols2grid-id"]).to_list())
    cell = driver.find_by_css_selector("#mols2grid .m2g-cell")
    data_el = cell.find_elements(By.CLASS_NAME, "data")
    classes = [
        c.replace("data-", "").replace("-display", "")
        for x in data_el
        for c in x.get_attribute("class").split(" ")
        if c.startswith("data-")
    ]
    classes = set(classes)
    assert classes == columns


def test_smiles_hidden(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-SMILES")
    assert not el.is_displayed()


@pytest.mark.parametrize("page", [1, 2, 3])
def test_page_click(driver: FirefoxDriver, grid, page):
    doc = get_doc(grid, dict(subset=["img", "_Name"], n_items_per_page=9))
    driver.get(doc)
    for i in range(2, page + 1):
        driver.wait_for_img_load()
        next_page = driver.find_by_css_selector(f'a.page-link[data-i="{i}"]')
        next_page.click()
    first_cell = driver.find_by_class_name("m2g-cell")
    mols2grid_id = 9 * (page - 1)
    name = first_cell.find_element(By.CLASS_NAME, "data-_Name")
    ref = grid.dataframe.iloc[mols2grid_id]
    assert name.text == ref["_Name"]


@pytest.mark.parametrize(
    ["name", "css_prop", "value", "expected"],
    [
        ("gap", "margin-top", 20, "20px"),
        ("border", "border-top-width", "3px solid", "3px"),
        ("border", "border-top-style", "1px dashed", "dashed"),
        ("border", "border-top-color", "1px solid blue", "rgb(0, 0, 255)"),
        ("fontsize", "font-size", "16pt", "21.3333px"),
        ("fontfamily", "font-family", "Consolas", "Consolas"),
        ("textalign", "text-align", "right", "right"),
        (
            "custom_css",
            "background-color",
            "#mols2grid .m2g-cell { background-color: black; }",
            "rgb(0, 0, 0)",
        ),
    ],
)
def test_css_properties(driver: FirefoxDriver, grid, name, css_prop, value, expected):
    doc = get_doc(grid, {name: value})
    driver.get(doc)
    computed = driver.execute_script(
        f"""return getComputedStyle(
            document.querySelector('#mols2grid .m2g-cell')
        ).getPropertyValue({css_prop!r});
        """
    )
    assert computed == expected


def test_text_search(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    driver.wait_for_img_load()
    driver.text_search("iodopropane")
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-SMILES")
    assert el.get_attribute("innerHTML") == "CC(I)C"


def test_text_search_regex_chars(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    driver.wait_for_img_load()
    driver.text_search("1-pentene")
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-SMILES")
    assert el.get_attribute("innerHTML") == "CCCC=C"


@flaky(max_runs=3, min_passes=1)
def test_smarts_search(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    driver.wait_for_img_load()
    driver.substructure_query("CC(I)C")
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-_Name")
    assert el.text == "2-iodopropane"


def test_selection_click(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    driver.wait_for_img_load()
    sel = driver.click_checkbox()
    assert sel == {0: "CCC(C)CC"}
    register._clear()


def test_export_csv(driver: FirefoxDriver, html_doc):
    sep = ";"
    driver.get(html_doc)
    driver.wait_for_img_load()
    driver.click_checkbox()
    driver.sort_grid("_Name")
    now = datetime.now(tz=timezone.utc)
    driver.grid_action("save-csv")
    csv_files = sorted(
        (Path.home() / "Downloads").glob("selection*.csv"),
        key=lambda x: x.stat().st_mtime,
    )
    csv_file = csv_files[-1]
    file_mtime = datetime.fromtimestamp(csv_file.stat().st_mtime, tz=timezone.utc)
    assert (file_mtime - now).seconds < 1, "Could not find recent selection file"
    content = csv_file.read_text()
    expected = (
        sep.join(("index", "_Name", "SMILES"))
        + "\n"
        + sep.join(("0", "3-methylpentane", "CCC(C)CC"))
        + "\n"
    )
    assert content == expected
    csv_file.unlink()
    register._clear()


def test_selection_with_cache_check_and_uncheck(driver: FirefoxDriver, df):
    register._init_grid("cached_sel")
    event = SimpleNamespace(new='{0: "CCC(C)CC"}')
    register.selection_updated("cached_sel", event)
    grid = get_grid(df, name="cached_sel", cache_selection=True)
    doc = get_doc(grid, {})
    driver.get(doc)
    driver.wait_for_img_load()
    sel = driver.wait_for_selection(is_empty=False)
    assert sel == {0: "CCC(C)CC"}
    empty_sel = driver.click_checkbox(is_empty=True)
    assert empty_sel is True
    register._clear()


def test_selection_check_uncheck_invert(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    driver.wait_for_img_load()
    # search
    driver.text_search("iodopropane")
    # check all
    driver.grid_action("select-all")
    sel = driver.wait_for_selection(is_empty=False)
    assert len(sel) == 30
    # uncheck all
    driver.grid_action("unselect-all")
    empty_sel = driver.wait_for_selection(is_empty=True)
    assert empty_sel is True
    # check matching
    driver.grid_action("select-matching")
    sel = driver.wait_for_selection(is_empty=False)
    assert sel == {27: "CC(I)C"}
    # invert
    driver.grid_action("invert")
    sel = driver.wait_for_selection(is_empty=False)
    assert len(sel) == 29
    register._clear()


@pytest.mark.parametrize("prerender", [True, False])
def test_image_size(driver: FirefoxDriver, df, prerender):
    size = (200, 300)
    grid = get_grid(df, size=size, prerender=prerender)
    doc = get_doc(
        grid,
        {
            "selection": False,
            "border": "0",
            "substruct_highlight": False,
            "n_items_per_page": 1,
        },
    )
    driver.get(doc)
    if not prerender:
        driver.wait_for_img_load()
    img = driver.find_by_css_selector("#mols2grid .m2g-cell .data-img *")
    assert img.size == {"width": size[0], "height": size[1]}


def test_image_use_coords(driver: FirefoxDriver, df):
    mols = df["mol"][:1]
    AllChem.EmbedMolecule(mols[0], randomSeed=0xF00D)
    grid = mols2grid.MolGrid.from_mols(
        mols,
        use_coords=True,
        prerender=True,
        useSVG=False,
        size=(160, 120),
    )
    doc = get_doc(grid, {"substruct_highlight": False})
    driver.get(doc)
    hash_ = driver.get_png_hash()
    diff = hash_ - imagehash.hex_to_hash(
        "ffffffffff7fff7ffe7ffe7ffe7ffe7ffe7fe07fc33f0fbc3f80ffc7ffffffff"
    )
    assert diff <= 2


@pytest.mark.parametrize(
    ["coordGen", "prerender", "expected"],
    [
        pytest.param(
            True,
            True,
            "fffffffffffffe7ffe7ffe7ffe7ffe7ffe7f3e7c8811c183e7e7ffffffffffff",
            marks=skip_no_coordgen,
        ),
        (
            True,
            False,
            "fffffffffffffe7ffe7ffe7ffe7ffe7ffe7f3e7c8811c183e7e7ffffffffffff",
        ),
        (
            False,
            True,
            "ffffffff03fcf9fcfdf9fcf9fcfbfe03fe07fcfffcfffdfff9fffbffffffffff",
        ),
        (
            False,
            False,
            "ffffffff03fcf9fcfdf9fcf9fcfbfe03fe07fcfffcfffdfff9fffbffffffffff",
        ),
    ],
)
def test_coordgen(driver: FirefoxDriver, mols, coordGen, prerender, expected):
    useSVG = not prerender
    grid = mols2grid.MolGrid.from_mols(
        mols,
        coordGen=coordGen,
        prerender=prerender,
        useSVG=useSVG,
        size=(160, 120),
        use_coords=False,
    )
    doc = get_doc(grid, {"substruct_highlight": False})
    driver.get(doc)
    if not prerender:
        driver.wait_for_img_load()
    hash_ = driver.get_svg_hash() if useSVG else driver.get_png_hash()
    assert str(hash_) == expected


@pytest.mark.parametrize(
    ["removeHs", "prerender", "expected"],
    [
        pytest.param(
            True,
            True,
            "fffffffffffffe7ffe7ffe7ffe7ffe7ffe7f3e7c8811c183e7e7ffffffffffff",
            marks=skip_no_coordgen,
        ),
        pytest.param(
            False,
            True,
            (
                "ffffff7fff3fffbff907e02fe13ff80fcbafe33fe033cb07fa4ffa4fff97ffd7"
                if rdkit_version == "2020.03.1"
                else "ff7ffe1ff91ffd3ff00ff0cffcbff0bff00ffd3fe1bff887f29ff30fff6fff7f"
            ),
            marks=skip_no_coordgen,
        ),
        (
            True,
            False,
            "fffffffffffffe7ffe7ffe7ffe7ffe7ffe7f3e7c8811c183e7e7ffffffffffff",
        ),
        (
            False,
            False,
            "ff7ffe1ff91ffd3ff00ff0cffcbff0bff00ffd3fe1bff887f29ff30fff6fff7f",
        ),
    ],
)
def test_removeHs(driver: FirefoxDriver, df, removeHs, prerender, expected):
    useSVG = not prerender
    mol = df["mol"][0]
    mol.ClearProp("SMILES")
    mols = [Chem.AddHs(mol)]
    grid = mols2grid.MolGrid.from_mols(
        mols,
        removeHs=removeHs,
        prerender=prerender,
        useSVG=useSVG,
        size=(160, 120),
        use_coords=False,
    )
    doc = get_doc(grid, {"n_items_per_page": 5, "substruct_highlight": False})
    driver.get(doc)
    if not prerender:
        driver.wait_for_img_load()
    if useSVG:
        hash_ = driver.get_svg_hash()
    else:
        hash_ = driver.get_png_hash()
    if expected == "":
        raise AssertionError(str(hash_))
    diff = hash_ - imagehash.hex_to_hash(expected)
    assert diff <= 1


@pytest.mark.parametrize(
    ["kwargs", "expected"],
    [
        (
            dict(addAtomIndices=True),
            "ffffffffff7ffe3ffe7ffefffeff3e7f3e791830c184e7cfe7cff7cfffffffff",
        ),
        (
            dict(fixedBondLength=10),
            "fffffffffffffffffffffffffe7ffe7ff81ffc3fffffffffffffffffffffffff",
        ),
        (
            dict(atomColourPalette={6: (0, 0.8, 0.8)}),
            "fffffffffffffffffe7ffe7ffe7ffe7ffe7f3e7c8819c183e7e7ffffffffffff",
        ),
        (
            dict(legend="foo"),
            "fffffffffffffe7ffe7ffe7ffe7ffe7f3e7c1818c183e7e7fffffffffe7ffe7f",
        ),
    ],
)
def test_moldrawoptions(driver: FirefoxDriver, df, kwargs, expected):
    grid = get_grid(df, size=(160, 120), **kwargs)
    doc = get_doc(grid, dict(n_items_per_page=1, subset=["img"]))
    driver.get(doc)
    driver.wait_for_img_load()
    hash_ = driver.get_svg_hash()
    assert str(hash_) == expected


@pytest.mark.xfail(GITHUB_ACTIONS, reason="only seem to pass locally")
def test_hover_color(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"hover_color": "rgba(255, 0, 0, 0.1)"})
    driver.get(doc)
    (
        ActionChains(driver)
        .move_to_element(driver.find_by_css_selector("#mols2grid .m2g-cell"))
        .pause(0.2)
        .perform()
    )
    color = driver.execute_script(
        """
        return getComputedStyle(
            document.querySelector('#mols2grid .m2g-cell:hover'), ":after"
        ).getPropertyValue('background-color');
        """
    )
    assert color == "rgba(255, 0, 0, 0.1)"


@flaky(max_runs=3, min_passes=1)
def test_tooltip(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"tooltip": ["_Name"]})
    driver.get(doc)
    driver.wait_for_img_load()
    tooltip = driver.get_tooltip_content()
    assert (
        tooltip
        == '<strong>_Name</strong>: <span class="copy-me">3-methylpentane</span>'
    )


@flaky(max_runs=3, min_passes=1)
def test_tooltip_fmt(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"tooltip": ["_Name"], "tooltip_fmt": "<em>{value}</em>"})
    driver.get(doc)
    driver.wait_for_img_load()
    tooltip = driver.get_tooltip_content()
    assert tooltip == '<em><span class="copy-me">3-methylpentane</span></em>'


def test_tooltip_not_in_subset(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"tooltip": ["_Name"], "subset": ["ID", "img"]})
    driver.get(doc)
    driver.wait_for_img_load()
    tooltip = driver.get_tooltip_content()
    assert (
        tooltip
        == '<strong>_Name</strong>: <span class="copy-me">3-methylpentane</span>'
    )


@flaky(max_runs=3, min_passes=1)
def test_style(driver: FirefoxDriver, grid):
    doc = get_doc(
        grid,
        {
            "tooltip": ["_Name"],
            "style": {
                "__all__": lambda x: "color: red",
                "_Name": lambda x: "color: blue",
            },
        },
    )
    driver.get(doc)
    driver.wait_for_img_load()
    el = driver.find_by_css_selector("#mols2grid .m2g-cell")
    assert el.value_of_css_property("color") == "rgb(255, 0, 0)"
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-_Name")
    assert el.value_of_css_property("color") == "rgb(0, 0, 255)"
    tooltip = driver.get_tooltip_content()
    assert (
        tooltip == '<strong>_Name</strong>: <span class="copy-me" '
        'style="color: blue">3-methylpentane</span>'
    )


@flaky(max_runs=3, min_passes=1)
def test_transform(driver: FirefoxDriver, grid):
    doc = get_doc(
        grid, {"tooltip": ["_Name"], "transform": {"_Name": lambda x: x.upper()}}
    )
    driver.get(doc)
    name = driver.find_by_css_selector("#mols2grid .m2g-cell .data-_Name")
    assert name.text == "3-METHYLPENTANE"
    driver.wait_for_img_load()
    tooltip = driver.get_tooltip_content(pause=0.5)
    assert (
        tooltip
        == '<strong>_Name</strong>: <span class="copy-me">3-METHYLPENTANE</span>'
    )


@flaky(max_runs=3, min_passes=1)
def test_transform_style_tooltip(driver: FirefoxDriver, grid):
    doc = get_doc(
        grid,
        {
            "tooltip": ["_Name"],
            "transform": {"_Name": lambda x: "foo"},
            "style": {
                "__all__": lambda x: "background-color: red",
                "_Name": lambda x: "color: green" if x == "foo" else "color: blue",
            },
        },
    )
    driver.get(doc)
    driver.wait_for_img_load()
    cell = driver.find_by_css_selector("#mols2grid .m2g-cell")
    assert cell.value_of_css_property("background-color") == "rgb(255, 0, 0)"
    name = cell.find_element(By.CLASS_NAME, "data-_Name")
    assert name.text == "foo"
    tooltip = driver.get_tooltip_content(pause=0.5)
    assert (
        tooltip == '<strong>_Name</strong>: <span class="copy-me" style="color: blue">'
        "foo</span>"
    )


@pytest.mark.parametrize("selection", [True, False])
def test_callback_js(driver: FirefoxDriver, grid, selection):
    doc = get_doc(
        grid,
        {
            "subset": ["img", "_Name"],
            "callback": "$('#mols2grid .m2g-cell .data-_Name').html('foo')",
            "selection": selection,
        },
    )
    driver.get(doc)
    driver.wait_for_img_load()
    driver.trigger_callback()
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-_Name")
    assert el.text == "foo"


def test_sort_by(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"subset": ["img", "_Name"], "sort_by": "_Name"})
    driver.get(doc)
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-_Name")
    assert el.text == "1,1,2,2-tetrachloroethane"


def test_sort_button(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    driver.wait_for_img_load()
    driver.sort_grid("_Name")
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-_Name")
    assert el.text == "1,1,2,2-tetrachloroethane"
    driver.invert_sort()
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-_Name")
    assert el.text == "tetrachloromethane"


@pytest.mark.parametrize(
    ["substruct_highlight", "expected"],
    [
        (True, "fffffe7ffc3ffc3ffe7ffe7ffe7ffe7ffe7ffc3ff81fc003c3c3c7e3c7e3eff7"),
        (False, "fe7ffe7ffe7ffe7ffe7ffe7ffe7ffe7ffe7ffe7ffe7ffc3ff99fe3c7c7e3cff3"),
    ],
)
def test_substruct_highlight(
    driver: FirefoxDriver, grid, substruct_highlight, expected
):
    doc = get_doc(
        grid, {"n_items_per_page": 5, "substruct_highlight": substruct_highlight}
    )
    driver.get(doc)
    driver.wait_for_img_load()
    driver.substructure_query("CC(I)C")
    hash_ = driver.get_svg_hash()
    assert str(hash_) == expected


def test_substruct_clear_removes_highlight(driver: FirefoxDriver, grid):
    doc = get_doc(
        grid,
        {
            "n_items_per_page": 5,
            "substruct_highlight": True,
            "size": (160, 120),
        },
    )
    driver.get(doc)
    driver.wait_for_img_load()
    driver.substructure_query("C")
    hash_hl = driver.get_svg_hash()
    driver.clear_search()
    hash_ = driver.get_svg_hash()
    assert hash_ != hash_hl
    assert (
        str(hash_) == "fffffffffffffe7ffe7ffe7ffe7ffe7ffe7f3e7c8811c183e7e7ffffffffffff"
    )


def test_smarts_to_text_search_removes_highlight(driver: FirefoxDriver, grid):
    doc = get_doc(
        grid,
        {
            "n_items_per_page": 5,
            "substruct_highlight": True,
            "size": (160, 120),
        },
    )
    driver.get(doc)
    driver.wait_for_img_load()
    driver.substructure_query("I")
    hash_hl = driver.get_svg_hash()
    driver.text_search("odopropane")
    hash_ = driver.get_svg_hash()
    assert hash_ != hash_hl
    assert (
        str(hash_) == "fe7ffe7ffe7ffe7ffe7ffe7ffe7ffe7ffe7ffe7ffe7ffc3ff99fe3c7c7e3cff3"
    )


def test_filter(driver: FirefoxDriver, grid):
    doc = get_doc(
        grid,
        {
            "subset": ["img", "_Name"],
        },
    )
    driver.get(doc)
    mask = grid.dataframe["_Name"].str.contains("iodopropane")
    filter_code = env.get_template("js/filter.js").render(
        grid_id=grid._grid_id, mask=json.dumps(mask.tolist())
    )
    driver.wait_for_img_load()
    driver.execute_script(filter_code)
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-_Name")
    assert el.text == "2-iodopropane"


def test_subset_gives_rows_order(driver: FirefoxDriver, grid):
    subset = ["_Name", "ID"]
    doc = get_doc(grid, {"subset": subset, "n_items_per_page": 5})
    driver.get(doc)
    driver.wait_for_img_load()
    cell = driver.find_by_css_selector("#mols2grid .m2g-cell")
    elements = cell.find_elements(By.CLASS_NAME, "data")
    i = 0
    for el in elements:
        class_list = el.get_attribute("class").split()
        name = [x.replace("data-", "") for x in class_list if x.startswith("data-")][0]
        if name in {"img", "SMILES"}:
            continue
        assert name == subset[i]
        i += 1
    # smiles should always be there, and last
    assert name == "SMILES"
    assert el.value_of_css_property("display") == "none"


def test_colname_with_spaces(driver: FirefoxDriver, df):
    df = df.rename(columns={"SMILES": "Molecule", "_Name": "Molecule name"}).drop(
        columns="mol"
    )
    grid = mols2grid.MolGrid(df, smiles_col="Molecule")
    doc = get_doc(
        grid,
        dict(subset=["Molecule name", "img"], tooltip=["Molecule"], n_items_per_page=5),
    )
    driver.get(doc)
    driver.wait_for_img_load()
    el = driver.find_by_css_selector("#mols2grid .m2g-cell .data-Molecule-name")
    assert el.text == "3-methylpentane"


@flaky(max_runs=3, min_passes=1)
def test_custom_header(driver: FirefoxDriver, grid):
    doc = get_doc(
        grid,
        {
            "subset": ["img"],
            "custom_header": '<script src="https://unpkg.com/@rdkit/rdkit@2021.9.2/Code/MinimalLib/dist/RDKit_minimal.js"></script>',
            "n_items_per_page": 5,
        },
    )
    driver.get(doc)
    driver.wait_for_img_load()
    val = driver.execute_script("return RDKit.version();")
    assert val == "2021.09.2"


def test_static_template(driver: FirefoxDriver, sdf_path):
    df = mols2grid.sdf_to_dataframe(sdf_path)[:15]
    grid = mols2grid.MolGrid(df, mol_col="mol", prerender=True, size=(160, 120))
    doc = get_doc(
        grid,
        dict(
            template="static",
            subset=["mols2grid-id", "img"],
            tooltip=["_Name"],
            sort_by="_Name",
            tooltip_trigger="hover",
        ),
    )
    driver.get(doc)
    el = driver.find_by_css_selector("#mols2grid td.col-0")
    assert el.find_element(By.CLASS_NAME, "data-mols2grid-id").text == "8"
    tooltip = driver.get_tooltip_content(selector=".m2g-cell-0")
    assert (
        tooltip
        == '<strong>_Name</strong>: <span class="copy-me">1,3,5-trimethylbenzene</span>'
    )
    if COORDGEN_SUPPORT:
        hash_ = driver.get_svg_hash("#mols2grid td .data-img")
        diff = hash_ - imagehash.hex_to_hash(
            "fffffe7ffe7ffe7ffe7ffe7ffc3ff10ff3cff3cff3cff3cff38fe1078c319e79"
        )
        assert diff <= 1


def test_default_subset_tooltip(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"n_items_per_page": 5})
    driver.get(doc)
    driver.wait_for_img_load()
    expected_subset = ["img"]
    expected_tooltip = [
        x
        for x in grid.dataframe.columns.drop(["mol", "mols2grid-id"]).to_list()
        if x not in expected_subset
    ]
    cell = driver.find_by_css_selector("#mols2grid .m2g-cell")
    data_elements = cell.find_elements(By.CLASS_NAME, "data")
    subset = [
        c.replace("data-", "").replace("-display", "")
        for x in data_elements
        for c in x.get_attribute("class").split(" ")
        if c.startswith("data-") and not x.get_attribute("style")
    ]
    assert subset == expected_subset
    tooltip = [
        c.replace("data-", "").replace("-display", "")
        for x in data_elements
        for c in x.get_attribute("class").split(" ")
        if c.startswith("data-") and x.get_attribute("style") == "display: none;"
    ]
    assert tooltip == expected_tooltip


def test_multi_highlight(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    driver.wait_for_img_load()
    driver.substructure_query("Br")
    hash_ = driver.get_svg_hash()
    assert (
        str(hash_) == "ffffffff8fff87ff003f079f8fdfffcfffefffe1ffe1ffe0ffe1fff1ffffffff"
    )


def test_single_highlight(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"single_highlight": True})
    driver.get(doc)
    driver.wait_for_img_load()
    driver.substructure_query("Br")
    hash_ = driver.get_svg_hash()
    assert (
        str(hash_) == "ffffffff8fff87ff003f001f87cfcfcfffefffe7fff3fff0fff1fff1ffffffff"
    )


def test_mol_depiction_aligned_to_query(driver: FirefoxDriver, html_doc):
    driver.get(html_doc)
    driver.wait_for_img_load()
    driver.substructure_query("CCCBr")
    images = list(driver.get_imgs_from_svgs())
    for img, expected in zip(
        images,
        [
            "fffffffffcf9f8f8f0708001070f0f8f9f9fffdfffdfff9fff8fff8fffffffff",
            "fffffffffffffff9fff9fff9fff9f8f9f8f0d0018003078f8f8fffffffffffff",
        ],
    ):
        hash_ = imagehash.average_hash(img, hash_size=16)
        assert str(hash_) == expected


def test_highlight_with_hydrogens(driver: FirefoxDriver, df):
    mols = []
    for mol in df["mol"][25:]:
        m = Chem.AddHs(mol)
        m.ClearProp("SMILES")
        mols.append(m)
    grid = mols2grid.MolGrid.from_mols(
        mols,
        removeHs=False,
        size=(160, 120),
    )
    doc = get_doc(grid, dict(substruct_highlight=True, single_highlight=False))
    driver.get(doc)
    driver.wait_for_img_load()
    driver.substructure_query("Cl")
    hash_ = driver.get_svg_hash()
    assert (
        str(hash_) == "fdfffcfff8fffcfffdc7cdc780cf885bd901fb81f3b3f3bfff9fff1fff9fffbf"
    )


def test_callbacks_info(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"callback": mols2grid.callbacks.info()})
    driver.get(doc)
    driver.trigger_callback()
    modal = driver.find_by_css_selector("#m2g-modal")
    assert (
        modal.find_element(By.CSS_SELECTOR, ".m2g-modal-header h2").get_attribute(
            "innerHTML"
        )
        == "CCC(C)CC"
    )
    content = modal.find_element(By.CSS_SELECTOR, ".m2g-modal-body").get_attribute(
        "innerHTML"
    )
    assert "PFEOZHBOMNWTJB-UHFFFAOYSA-N" in content


def test_callbacks_3D(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"callback": mols2grid.callbacks.show_3d()})
    driver.get(doc)
    driver.trigger_callback(pause=2.0)
    modal = driver.find_by_css_selector("#m2g-modal")
    assert (
        modal.find_element(By.CSS_SELECTOR, ".m2g-modal-header h2").get_attribute(
            "innerHTML"
        )
        == "CCC(C)CC"
    )
    content = modal.find_element(By.CSS_SELECTOR, ".m2g-modal-body").get_attribute(
        "innerHTML"
    )
    assert '<div id="molviewer' in content
    if not GITHUB_ACTIONS:
        # only works when testing locally...
        assert "<canvas" in content
        # cannot test for actual rendering as there's no GL available
        assert driver.execute_script("return typeof($3Dmol)") != "undefined"


def test_callbacks_external_link(driver: FirefoxDriver, grid):
    doc = get_doc(grid, {"callback": mols2grid.callbacks.external_link()})
    driver.get(doc)
    driver.trigger_callback()
    # check if new tab was opened
    assert len(driver.window_handles) > 1
    urls = []
    for handle in driver.window_handles[1:]:
        driver.switch_to.window(handle)
        driver.wait(EC.url_contains("https://"))
        url = driver.current_url
        if url == "https://leruli.com/search/Q0NDKEMpQ0M=/home":
            break
        urls.append(url)
    else:
        raise AssertionError("Corresponding URL not found", urls)



from typing import NamedTuple, Optional

from .utils import env


class _JSCallback(NamedTuple):
    """Class that holds JavaScript code for running a callback function. If an external
    library is required for the callback to function correctly, it can be passed in
    the optional ``library_src`` as a ``<script>`` tag.
    """

    code: str
    library_src: Optional[str] = None


def make_popup_callback(title=None, subtitle=None, svg=None, html="", js="", style=""):
    """Creates a JavaScript callback that displays a popup window

    Parameters
    ----------
    title : str
        Title of the popup. Use ``title='${data["Name"]}'`` to use the value
        of the column "Name" as a title
    subtitle : str
        Secondary title which works the same as title.
    svg : str
        SVG depiction of the molecule
    html : str
        Content of the popup window
    js : str
        JavaScript code executed before making the content of the popup window.
        This allows you to create variables and reuse them later in the `html`
        content of the popup, using the ``${my_variable}`` syntax
    style : str
        CSS style assigned to the popup window

    Returns
    -------
    js_callback : str
        JavaScript code that allows to display a popup window
    """
    return env.get_template("js/popup.js").render(
        title=title, subtitle=subtitle, html=html, svg=svg, js=js, style=style
    )


def _get_title_field(title):
    return "${data['" + title + "']}" if title else None


def info(title="SMILES", subtitle=None, img_size=(400, 300), style="") -> _JSCallback:
    """Displays a bigger image of the molecule, alongside some useful descriptors:
    molecular weight, number of Hydrogen bond donors and acceptors, TPSA, Crippen ClogP
    and InChIKey.

    Parameters
    ----------
    title : str
        Title of the popup. Use ``title='${data["Name"]}'`` to use the value
        of the column "Name" as a title
    subtitle : str
        Secondary title which works the same as title.
    img_size : tuple
        Width and height of the molecule depiction.
    style : str
        CSS style applied to the modal window.
    """
    code = make_popup_callback(
        title=_get_title_field(title),
        subtitle=_get_title_field(subtitle),
        svg="${svg}",
        js=f"""
            let mol = RDKit.get_mol(data["SMILES"]);
            let svg = mol.get_svg({img_size[0]}, {img_size[1]});
            let desc = JSON.parse(mol.get_descriptors());
            let inchikey = RDKit.get_inchikey_for_inchi(mol.get_inchi());
            mol.delete();
        """,
        html="""
            <b>Molecular weight</b>: ${desc.exactmw}<br/>
            <b>HBond Acceptors</b>: ${desc.NumHBA}<br/>
            <b>HBond Donors</b>: ${desc.NumHBD}<br/>
            <b>TPSA</b>: ${desc.tpsa}<br/>
            <b>ClogP</b>: ${desc.CrippenClogP}<br/>
            <hr>
            <b>InChIKey</b>: ${inchikey}
            """,
        style=style,
    )
    return _JSCallback(code=code)


def show_3d(
    title="SMILES",
    subtitle=None,
    query=["pubchem", "cactus"],
    height="100%",
    style="width:100%;height:100%",
) -> _JSCallback:
    """Queries the API(s) listed in ``query`` using the SMILES of the structure, to
    fetch the 3D structure and display it with ``3Dmol.js``

    Parameters
    ----------
    title : str
        Title of the popup. Use ``title='${data["Name"]}'`` to use the value
        of the column "Name" as a title
    subtitle : str
        Secondary title which works the same as title.
    query : list or dict
        List of APIs used to fetch the 3D structure from (by order of priority). To use
        a custom API, use a dict with the following format::

            {
                "url": "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/smiles/{}/SDF?record_type=3d",
                "format": "sdf",
                "field": "SMILES",
                "encode": True,
            }

        In this example, the value in the ``SMILES`` field will be URI-encoded and
        replace the curly-braces in the ``url`` strings to create the URL to fetch
        the 3D structure from in the SDF format.
    height : str
        Height of the 3Dmol.js viewer in the modal.
    style : str
        CSS style applied to the modal window.
    """
    js_script = env.get_template("js/callbacks/show_3d.js").render(query=query)
    code = make_popup_callback(
        title=_get_title_field(title),
        subtitle=_get_title_field(subtitle),
        html=f'<div id="molviewer" style="width: 100%; height: {height};"></div>',
        js=js_script,
        style=style,
    )
    library = """<script src="https://cdnjs.cloudflare.com/ajax/libs/3Dmol/1.8.0/3Dmol-nojquery-min.js" integrity="sha512-9iiTgstim185ZZPL/nZ+t+MLMmIbZEMfoZ1swSBUhxt4AukOPY34yyO2217X1dN5ziVMKi+YLmp/JBj+KyEaUQ==" crossorigin="anonymous" referrerpolicy="no-referrer"></script>"""
    return _JSCallback(code=code, library_src=library)


def external_link(
    url="https://leruli.com/search/{}/home",
    field="SMILES",
    url_encode=False,
    b64_encode=True,
) -> _JSCallback:
    """Opens an external link using ``url`` as a template string and the value in the
    corresponding ``field``. The value can be URL-encoded or base64-encoded if needed.

    Parameters
    ----------
    url : str
        Template string used to generate the URL that will be opened.
    field : str
        Field name used to generate the URL that will be opened. The value can be
        encoded (see below).
    url_encode : bool
        Encode the value fetched from the specified field to replace characters that are
        not allowed in a URL e.g., spaces become ``%20``.
    b64_encode : bool
        Base64-encode the value fetched from the field.

    Raises
    ------
    ValueError : Both ``url_encode`` and ``b64_encode`` have been specified.
    """
    if url_encode and b64_encode:
        raise ValueError("Setting both URL and B64 encoding is not supported")
    code = env.get_template("js/callbacks/external_link.js").render(
        url=url,
        field=field,
        url_encode=url_encode,
        b64_encode=b64_encode,
    )
    return _JSCallback(code=code)



import gzip
import re
from ast import literal_eval
from functools import partial, wraps
from importlib.util import find_spec
from pathlib import Path

import pandas as pd
from jinja2 import Environment, FileSystemLoader
from rdkit import Chem

env = Environment(
    loader=FileSystemLoader(Path(__file__).parent / "templates"), autoescape=False
)


def requires(module):
    def inner(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if find_spec(module):
                return func(*args, **kwargs)
            raise ModuleNotFoundError(
                f"The module {module!r} is required to use {func.__name__!r} "
                "but it is not installed!"
            )

        return wrapper

    return inner


def tooltip_formatter(s, subset, fmt, style, transform):
    """Function to generate tooltips from a pandas Series

    Parameters
    ----------
    s : pandas.Series
        Row in the internal pandas DataFrame
    subset : list
        Subset of columns that are used for the tooltip
    fmt : str
        Format string for each key-value pair of the tooltip
    style : dict
        CSS styling applied to each item independently
    transform : dict
        Functions applied to each value before rendering
    """
    items = []
    for k, v in s[subset].to_dict().items():
        displayed = transform[k](v) if transform.get(k) else v
        v = (
            f'<span class="copy-me" style="{style[k](v)}">{displayed}</span>'
            if style.get(k)
            else f'<span class="copy-me">{displayed}</span>'
        )
        items.append(fmt.format(key=k, value=v))
    return '<br>'.join(items)


def mol_to_smiles(mol):
    """Returns a SMILES from an RDKit molecule, or None if not an RDKit mol"""
    return Chem.MolToSmiles(mol) if mol else None


def mol_to_record(mol, mol_col="mol"):
    """Function to create a dict of data from an RDKit molecule"""
    return {**mol.GetPropsAsDict(includePrivate=True), mol_col: mol} if mol else {}


def sdf_to_dataframe(sdf_path, mol_col="mol"):
    """Creates a dataframe of molecules from an SDFile. All property fields in
    the SDFile are made available in the resulting dataframe

    Parameters
    ----------
    sdf_path : str, Path
        Path to the SDFile, ending with either ``.sdf`` or ``.sdf.gz``
    mol_col : str
        Name of the column containing the RDKit molecules in the dataframe

    Returns
    -------
    df : pandas.DataFrame
    """
    if str(sdf_path).endswith(".gz"):
        read_file = gzip.open
    else:
        read_file = partial(open, mode="rb")
    with read_file(sdf_path) as f:
        return pd.DataFrame(
            [mol_to_record(mol, mol_col)
             for mol in Chem.ForwardSDMolSupplier(f)]
        )


def remove_coordinates(mol):
    """Removes the existing coordinates from the molecule. The molecule is
    modified inplace"""
    mol.RemoveAllConformers()
    return mol


def slugify(string):
    """Replaces whitespaces with hyphens"""
    return re.sub(r"\s+", "-", string)


def callback_handler(callback, event):
    """Handler for applying the callback function on change"""
    data = literal_eval(event.new)
    callback(data)


def _get_streamlit_script_run_ctx():
    from streamlit.runtime.scriptrunner import get_script_run_ctx

    return get_script_run_ctx()


def is_running_within_streamlit():
    """
    Function to check whether python code is run within streamlit

    Returns
    -------
    use_streamlit : boolean
        True if code is run within streamlit, else False
    """
    try:
        ctx = _get_streamlit_script_run_ctx()
    except ImportError:
        return False
    else:
        return ctx is not None



from ._version import __version__
from .callbacks import make_popup_callback
from .dispatch import display, save
from .molgrid import MolGrid
from .select import get_selection, list_grids
from .utils import sdf_to_dataframe
from .widget import _jupyter_labextension_paths, _jupyter_nbextension_paths

try:
    from google.colab import output
except ImportError:
    pass
else:
    output.enable_custom_widget_manager()



import json
import warnings
from base64 import b64encode
from functools import partial
from html import escape

import numpy as np
import pandas as pd
from rdkit import Chem
from rdkit.Chem import Draw

from .callbacks import _JSCallback
from .select import register
from .utils import (
    callback_handler,
    env,
    is_running_within_streamlit,
    mol_to_record,
    mol_to_smiles,
    remove_coordinates,
    requires,
    sdf_to_dataframe,
    slugify,
    tooltip_formatter,
)
from .widget import MolGridWidget

try:
    from IPython.display import HTML, Javascript, display
except ModuleNotFoundError:
    pass
else:
    warnings.filterwarnings("ignore", "Consider using IPython.display.IFrame instead.")


# Detect if mols2grid is running inside a Jupyter Notebook/Lab.
# If it is, we wrap the HTML in an iframe.
try:
    get_ipython()  # This is callable only in Jupyter Notebook.
    is_jupyter = True
except NameError:
    is_jupyter = False


class MolGrid:
    """Class that handles drawing molecules, rendering the HTML document and
    saving or displaying it in a Jupyter Notebook.

    Parameters: Data
    ----------------
    df : pandas.DataFrame, dict or list, required
        Dataframe containing a SMILES or mol column, or dictionary containing a
        list of SMILES, or list of dictionnaries containing a SMILES field.
    smiles_col : str or None, default="SMILES"
        Name of the SMILES column in the dataframe, if available.
    mol_col : str or None, default=None
        Name of an RDKit molecule column. If available, coordinates and
        atom/bonds annotations from this will be used for depiction.

    Parameters: Display
    -------------------
    size : tuple, default=(130, 90)
        The size of the drawing canvas. The cell minimum width is set to the
        width of the image, so if the cell padding is increased, the image will
        be displayed smaller.
    useSVG : bool, default=True
        Use SVG images instead of PNG.
    prerender : bool, default=False
        Prerender images for the entire dataset, or generate them on-the-fly.
        Prerendering is slow and memory-hungry, but required when ``template="static"``
        or ``useSVG=False``.
    cache_selection : bool, default=False
        Restores the selection from a previous grid with the same name.

    Parameters: Mols
    ----------------
    removeHs : bool, default=False
        Remove hydrogen atoms from the drawings.
    use_coords : bool, default=False
        Use the coordinates of the molecules (only relevant when an SDF file, a
        list of molecules or a DataFrame of RDKit molecules were used as input.)
    coordGen : bool, default=True
        Use the CoordGen library instead of the RDKit one to depict the
        molecules in 2D.
    MolDrawOptions : rdkit.Chem.Draw.rdMolDraw2D.MolDrawOptions or None, default=None
        Drawing options. Useful for making highly customized drawings.

    Parameters: Customization
    -------------------------
    name : str, default="default"
        Name of the grid. Used when retrieving selections from multiple grids
        at the same time.
    rename : dict or None, default=None
        Rename the properties in the final document.
    kwargs : object
        :class:`~rdkit.Chem.Draw.rdMolDraw2D.MolDrawOptions` attributes, and
        the additional ``atomColourPalette``.

    Notes
    -----
    On-the-fly rendering of images does not read the atom colour palette
    from the :class:`~rdkit.Chem.Draw.rdMolDraw2D.MolDrawOptions` parameter.
    If this is needed, use the following::

        MolGrid(df, atomColourPalette={1: (.8, 0, 1)})

    .. versionchanged:: 0.1.0
        Added ``rename`` parameter to replace ``mapping``.

    .. versionadded:: 0.2.0
        Added ``prerender`` and ``cache_selection`` parameters.

    .. versionchanged:: 0.2.0
        Images are now generated on-the-fly. ``use_coords`` is now ``False`` by
        default to avoid a systematic error when using ``MolGrid.from_sdf``.
    """

    def __init__(
        self,
        df,
        smiles_col="SMILES",
        mol_col=None,
        #
        size=(130, 90),
        useSVG=True,
        prerender=False,
        cache_selection=False,
        #
        removeHs=False,
        use_coords=False,
        coordGen=True,
        MolDrawOptions=None,
        #
        name="default",
        rename=None,
        **kwargs,
    ):
        if not (smiles_col or mol_col):
            raise ValueError("One of `smiles_col` or `mol_col` must be set")
        if not isinstance(name, str):
            raise TypeError(
                f"`name` must be a string. Currently of type {type(name).__name__}"
            )
        if not prerender:
            if not useSVG:
                raise ValueError("On-the-fly rendering of PNG images not supported")
            if use_coords and mol_col:
                raise ValueError("Cannot use coordinates with on-the-fly rendering")
        self.prefer_coordGen = coordGen
        self.removeHs = removeHs
        self.useSVG = useSVG
        self.use_coords = use_coords
        self.img_size = size
        self.prerender = prerender
        self.smiles_col = smiles_col
        self.mol_col = mol_col
        if isinstance(df, pd.DataFrame):
            dataframe = df.copy()
        else:
            # List of dicts or other input formats for dataframes.
            dataframe = pd.DataFrame(df)
        if rename:
            dataframe.rename(columns=rename, inplace=True)
        self._extra_columns = ["img", "mols2grid-id"]
        # Add index.
        dataframe["mols2grid-id"] = list(range(len(dataframe)))
        # Generate drawing options.
        if prerender:
            Draw.rdDepictor.SetPreferCoordGen(coordGen)
            opts = MolDrawOptions or Draw.MolDrawOptions()
            for key, value in kwargs.items():
                setattr(opts, key, value)
            self.MolDrawOptions = opts
            self._MolDraw2D = Draw.MolDraw2DSVG if useSVG else Draw.MolDraw2DCairo
        else:
            opts = {}
            if MolDrawOptions:
                for key in dir(MolDrawOptions):
                    value = getattr(MolDrawOptions, key)
                    if not (
                        key.startswith("_")
                        or callable(value)
                        or value.__class__.__module__ != "builtins"
                    ):
                        opts[key] = value
            opts.update(kwargs)
            opts.update({"width": self.img_size[0], "height": self.img_size[1]})
            self.json_draw_opts = json.dumps(opts)
        # Prepare smiles and images.
        self._prepare_dataframe(dataframe)
        self.dataframe = dataframe
        # Register instance.
        self._grid_id = name
        if cache_selection:
            try:
                self._cached_selection = register.get_selection(name)
            except KeyError:
                self._cached_selection = {}
                register._init_grid(name)
            else:
                register._update_current_grid(name)
        else:
            self._cached_selection = {}
            register._init_grid(name)
        # Create widget.
        widget = MolGridWidget(grid_id=name, selection=str(self._cached_selection))
        selection_handler = partial(register.selection_updated, name)
        widget.observe(selection_handler, names=["selection"])
        # Register widget JS-side.
        display(widget)
        self.widget = widget

    @classmethod
    def from_mols(cls, mols, **kwargs):
        """Set up the dataframe used by mols2grid directly from a list of RDKit
        molecules.

        Parameters
        ----------
        mols : list
            List of RDKit molecules
        kwargs : object
            Other arguments passed on initialization
        """
        mol_col = kwargs.pop("mol_col", "mol")
        df = pd.DataFrame([mol_to_record(mol, mol_col=mol_col) for mol in mols])
        return cls(df, mol_col=mol_col, **kwargs)

    @classmethod
    def from_sdf(cls, sdf_file, **kwargs):
        """Set up the dataframe used by mols2grid directly from an SDFile.

        Parameters
        ----------
        sdf_file : str, pathlib.Path
            Path to the SDF file (.sdf or .sdf.gz)
        kwargs : object
            Other arguments passed on initialization


        .. versionchanged:: 0.2.0
            Added support for `.sdf.gz` files
        """
        mol_col = kwargs.pop("mol_col", "mol")
        df = sdf_to_dataframe(sdf_file, mol_col=mol_col)
        return cls(df, mol_col=mol_col, **kwargs)

    @property
    def template(self):
        """Kind of grid displayed, one of:

        * interactive
        * static
        """
        return self._template

    @template.setter
    def template(self, value):
        if value not in ["interactive", "static"]:
            raise ValueError(
                f"template={value!r} not supported. "
                "Use either 'interactive' or 'static'."
            )
        self._template = value

    def draw_mol(self, mol):
        """Draw a molecule."""
        d2d = self._MolDraw2D(*self.img_size)
        d2d.SetDrawOptions(self.MolDrawOptions)
        hl_atoms = getattr(mol, "__sssAtoms", [])
        d2d.DrawMolecule(mol, highlightAtoms=hl_atoms)
        d2d.FinishDrawing()
        return d2d.GetDrawingText()

    def mol_to_img(self, mol):
        """Convert an RDKit mol to an inline PNG image containing a drawing of the
        molecule."""
        img = self.draw_mol(mol)
        if self.useSVG:
            return img
        data = b64encode(img).decode()
        return f'<img src="data:image/png;base64,{data}">'

    def _prepare_dataframe(self, dataframe):
        """Prepares the dataframe with SMILES and images depending on user input. The
        dataframe is modified inplace."""
        if self.prerender:
            if self.mol_col:
                keep_mols = True
            else:
                # Make temporary mol column if not present.
                self.mol_col = "mol"
                keep_mols = False
                dataframe[self.mol_col] = dataframe[self.smiles_col].apply(
                    Chem.MolFromSmiles
                )
            # Drop empty mols.
            dataframe.dropna(axis=0, subset=[self.mol_col], inplace=True)
            # Modify mol according to user pref.
            if not self.use_coords:
                dataframe[self.mol_col] = dataframe[self.mol_col].apply(
                    remove_coordinates
                )
            if self.removeHs:
                dataframe[self.mol_col] = dataframe[self.mol_col].apply(Chem.RemoveHs)
            # Render.
            dataframe["img"] = dataframe[self.mol_col].apply(self.mol_to_img)
            # Cleanup.
            if not keep_mols:
                dataframe.drop(columns=self.mol_col, inplace=True)
                self.mol_col = None
        else:
            dataframe["img"] = None
        # Generate smiles col if not present or needs to be updated.
        if self.mol_col and self.smiles_col not in dataframe.columns:
            dataframe[self.smiles_col] = dataframe[self.mol_col].apply(mol_to_smiles)

    def render(self, template="interactive", **kwargs):
        """Returns the HTML document corresponding to the "interactive" or "static"
        template. See :meth:`to_interactive` and :meth:`to_static` for the full list
        of arguments.

        Parameters
        ----------
        template : str
            What kind of grid to draw:

            * interactive
                An interactive grid that layouts the original set of
                molecules on several pages, allowing for selecting molecules and
                filtering them using text or substructure queries.
            * static
                A simple table with all molecules displayed at once, similarly
                to RDKit's :func:`~rdkit.Chem.Draw.rdMolDraw2D.MolsToGridImage`.
                This template is mainly used for printing on paper or in a PDF
                file. Most of the interactive actions aren't available.
        """
        self.template = template
        return getattr(self, f"to_{self.template}")(**kwargs)

    def to_interactive(
        self,
        # Display
        subset=None,
        tooltip=None,
        tooltip_fmt="<strong>{key}</strong>: {value}",
        tooltip_placement="auto",
        transform=None,
        sort_by=None,
        use_iframe=False,
        truncate=True,
        n_items_per_page=24,
        selection=True,
        # Mols
        substruct_highlight=None,
        single_highlight=False,
        # CSS
        border="1px solid #cccccc",
        gap=0,
        pad=10,
        fontsize="12px",
        fontfamily="'DejaVu', sans-serif",
        textalign="center",
        background_color="white",
        hover_color="rgba(0,0,0,0.05)",
        custom_css=None,
        style=None,
        # Customization
        custom_header=None,
        callback=None,
        **kwargs,
    ):
        """Returns the HTML document for the "interactive" template.

        Parameters: Display
        -------------------
        subset: list or None, default=None
            Columns to be displayed in each cell of the grid. Each column's
            value will be displayed from top to bottom in the order provided.
            The ``"img"`` and ``"mols2grid-id"`` columns are displayed by default,
            however you can still add the ``"img"`` column if you wish to change
            the display order.
        tooltip : list, None or False, default=None
            Columns to be displayed inside the tooltip. When no subset is set,
            all columns will be listed in the tooltip by default. Use ``False``
            to hide the tooltip.
        tooltip_fmt : str, default="<strong>{key}</strong>: {value}"
            Format string of each key/value pair in the tooltip.
        tooltip_placement : str, default="auto"
            Position of the tooltip: ``auto``, ``top``, ``bottom``, ``left`` or
            ``right``.
        transform : dict or None, default=None
            Functions applied to specific items in all cells. The dict must follow
            a ``key: function`` structure where the key must correspond to one of
            the columns in ``subset`` or ``tooltip``. The function takes the item's
            value as input and transforms it, for example::

                transform={
                    "Solubility": lambda x: f"{x:.2f}",
                    "Melting point": lambda x: f"MP: {5/9*(x-32):.1f}°C"
                }

            These transformations only affect columns in ``subset`` and
            ``tooltip``, and do not interfere with ``style``.
        sort_by : str or None, default=None
            Sort the grid according to the following field (which must be
            present in ``subset`` or ``tooltip``).
        use_iframe : bool, default=False
            Whether to use an iframe to display the grid. When the grid is displayed
            inside a Jupyter Notebook or JupyterLab, this will default to ``True``.
        truncate: bool, default=True/False
            Whether to truncate the text in each cell if it's too long.
            Defaults to ``True`` for interactive grids, ``False`` for static grid.
        n_items_per_page, default=24
            Number of items to display per page. A multiple of 12 is recommended
            for optimal display.
        selection : bool, default=True
            Enables the selection of molecules and displays a checkbox at the
            top of each cell. In the context of a Jupyter Notebook, this gives
            you access to your selection (index and SMILES) through
            :func:`mols2grid.get_selection()` or :meth:`MolGrid.get_selection()`.
            In all cases, you can export your selection by clicking on the triple-dot menu.

        Parameters: Mols
        ----------------
        substruct_highlight : bool or None, default=None
            Highlight substructure when using the SMARTS search. Active by default
            when ``prerender=False``.
        single_highlight : bool, default=False
            Highlight only the first match of the substructure query.

        Parameters: CSS
        ---------------
        border : str, default="1px solid #cccccc"
            Styling of the border around each cell.
        gap : int, default=0
            Size in pixels of the gap between cells.
        pad : int, default=10
            Size in pixels of the cell padding.
        fontsize : str, default="12px"
            Font size of the text displayed in each cell.
        fontfamily : str, default="'DejaVu', sans-serif"
            Font used for the text in each cell.
        textalign : str, default="center"
            Alignment of the text in each cell.
        background_color : str, default="white"
            Background color of a cell.
        hover_color : str, default="rgba(0,0,0,0.05)"
            Background color when hovering a cell.
        custom_css : str or None, default=None
            Custom CSS properties applied to the generated HTML. Please note that
            the CSS will apply to the entire page if no iframe is used (see
            ``use_iframe`` for more details).
        style : dict or None, default=None
            CSS styling applied to each item in a cell. The dict must follow a
            ``key: function`` structure where the key must correspond to one of the
            columns in ``subset`` or ``tooltip``. The function takes the item's
            value as input, and outputs a valid CSS styling. For example, if you
            want to color the text corresponding to the "Solubility" column in your
            dataframe::

                style={"Solubility": lambda x: "color: red" if x < -5 else ""}

            You can also style a whole cell using the ``__all__`` key, the
            corresponding function then has access to all values for each cell::

                style={"__all__": lambda x: "color: red" if x["Solubility"] < -5 else ""}

            Parameters: Customization
            -------------------------
            custom_header : str or None, default=None
                Custom libraries to be loaded in the header of the document.
            callback : str, callable or None, default=None
                JavaScript or Python callback to be executed when clicking on an image.
                A dictionnary containing the data for the full cell is directly available
                as ``data`` in JS. For Python, the callback function must have ``data``
                as the first argument to the function. All the values in the ``data`` dict
                are parsed as strings, except "mols2grid-id" which is always an integer.
                Note that fields containing spaces in their name will be replaced by
                hyphens, i.e. "mol weight" becomes available as ``data["mol-weight"]``.

        Returns
        -------
        html_document : str

        Notes
        -----
        If ``subset=None, tooltip=None``, the index and image will be directly
        displayed on the grid while the remaining fields will be in the
        tooltip.

        The cell width is defined by the size[0] parameter.

        .. versionadded:: 0.1.0
            Added ``sort_by``, ``custom_css``, ``custom_header`` and
            ``callback`` arguments.
            Added the ability to style an entire cell with
            ``style={"__all__": <function>}``.

        .. versionadded:: 0.2.0
            Added ``substruct_highlight`` argument.

        .. versionchanged:: 0.2.2
            If both ``subset`` and ``tooltip`` are ``None``, the index and
            image will be directly displayed on the grid while the remaining
            fields will be in the tooltip.

        .. versionchanged:: 1.0.0
            ``callback`` can now be a *lambda* function. If ``prerender=True``,
            substructure highlighting will be automatically disabled if it
            wasn't explicitely set to ``True`` instead of raising an error.

        """
        if substruct_highlight is None:
            substruct_highlight = not self.prerender
        if substruct_highlight and self.prerender:
            raise ValueError(
                "Cannot highlight substructure search with prerendered images"
            )
        if self.mol_col:
            df = self.dataframe.drop(columns=self.mol_col).copy()
        else:
            df = self.dataframe.copy()
        smiles = self.smiles_col
        content = []  # Gets filled with the HTML content of each cell.
        column_map = {}

        if subset is None:
            if tooltip is None:
                subset = ["mols2grid-id", "img"]
                tooltip = [x for x in df.columns.tolist() if x not in subset]
            else:
                # When no subset is defined, all columns are displayed.
                subset = df.columns.tolist()
        else:
            # work on a copy
            subset = subset[:]

        if "mols2grid-id" not in subset:
            subset.insert(0, "mols2grid-id")
        if "img" not in subset:
            subset.insert(0, "img")

        # Always make sure the image comes first.
        # subset = [subset.pop(subset.index("img"))] + subset
        #
        # This was removed at Cedric's request, so you can choose
        # to have certain properties displayed above the image.

        # Define fields that are searchable and sortable.
        search_cols = [f"data-{col}" for col in subset if col != "img"]
        if tooltip:
            search_cols.extend([f"data-{col}" for col in tooltip])
            for col in tooltip:
                if col not in subset:
                    s = f'<div class="data data-{slugify(col)}" style="display: none;"></div>'
                    content.append(s)
                    column_map[col] = f"data-{col}"
        else:
            tooltip = []
        sort_cols = search_cols[:]
        sort_cols = ["data-mols2grid-id"] + sort_cols

        # Get unique list but keep order.
        sort_cols = list(dict.fromkeys(sort_cols))
        if style is None:
            style = {}
        if transform is None:
            transform = {}
        value_names = list(set(subset + [smiles] + tooltip))
        value_names = [f"data-{col}" for col in value_names]

        # Force id, SMILES, and tooltip values to be present in the data.
        final_columns = subset[:]
        final_columns.extend(["mols2grid-id", smiles])
        if tooltip:
            final_columns.extend(tooltip)
        final_columns = list(set(final_columns))

        # Make a copy of id shown explicitly.
        id_name = "mols2grid-id-display"
        df[id_name] = df["mols2grid-id"]
        value_names.append(f"data-{id_name}")
        final_columns.append(id_name)
        subset = [id_name if x == "mols2grid-id" else x for x in subset]
        id_display_html = f'<div class="data-{id_name}"></div>'

        # Organize data.
        temp = []
        for col in subset:
            if col == "mols2grid-id-display":
                s = ""  # Avoid an empty div to be created for the display id.
            elif col == "img" and tooltip:
                s = f'<a class="data data-{col}"></a>'
            else:
                if style.get(col):
                    s = (
                        f'<div class="data data-{slugify(col)} copy-me style-{slugify(col)}" '
                        'style=""></div>'
                    )
                else:
                    s = f'<div class="data data-{slugify(col)} copy-me"></div>'
            temp.append(s)
            column_map[col] = f"data-{col}"
        content = temp + content

        # Add but hide SMILES div if not present.
        if smiles not in (subset + tooltip):
            s = f'<div class="data data-{slugify(smiles)} copy-me" style="display: none;"></div>'
            content.append(s)
            column_map[smiles] = f"data-{smiles}"

        # Set mapping for list.js.
        if "__all__" in style.keys():
            whole_cell_style = True
            x = "[{data: ['mols2grid-id', 'cellstyle']}, "
        else:
            whole_cell_style = False
            x = "[{data: ['mols2grid-id']}, "
        value_names = [slugify(c) for c in value_names]
        value_names = x + str(value_names)[1:]

        # Apply CSS styles.
        for col, func in style.items():
            if col == "__all__":
                name = "cellstyle"
                df[name] = df.apply(func, axis=1)
            else:
                name = f"style-{slugify(col)}"
                df[name] = df[col].apply(func)
            final_columns.append(name)
            value_names = value_names[:-1] + f", {{ attr: 'style', name: {name!r} }}]"

        # Create tooltip.
        if tooltip:
            df["m2g-tooltip"] = df.apply(
                tooltip_formatter, axis=1, args=(tooltip, tooltip_fmt, style, transform)
            )
            final_columns += ["m2g-tooltip"]
            value_names = (
                value_names[:-1] + ", {attr: 'data-content', name: 'm2g-tooltip'}]"
            )
            info_btn_html = '<div class="m2g-info">i</div>'
        else:
            info_btn_html = ""

        # Apply custom user function.
        for col, func in transform.items():
            df[col] = df[col].apply(func)

        # Add checkboxes.
        if selection:
            if self._cached_selection:
                df["cached_checkbox"] = False
                df.loc[
                    df["mols2grid-id"].isin(self._cached_selection.keys()),
                    "cached_checkbox",
                ] = True
                final_columns += ["cached_checkbox"]
                value_names = (
                    value_names[:-1] + ", {attr: 'checked', name: 'cached_checkbox'}]"
                )
            checkbox_html = (
                '<input type="checkbox" tabindex="-1" '
                'class="position-relative float-left cached_checkbox">'
            )
        else:
            checkbox_html = ""

        # Add callback button.
        if callback:
            callback_btn_html = '<div class="m2g-callback"></div>'
        else:
            callback_btn_html = ""

        # Generate cell HTML.
        item = (
            '<div class="m2g-cell" data-mols2grid-id="0" tabindex="0">'
            '<div class="m2g-cb-wrap">{checkbox_html}<div class="m2g-cb"></div>{id_display_html}</div>'
            '<div class="m2g-cell-actions">{info_btn_html}{callback_btn_html}</div>'
            "{content}"
            "{tooltip_html}"
            "</div>"
        )

        item = item.format(
            checkbox_html=checkbox_html,
            id_display_html=id_display_html,
            info_btn_html=info_btn_html,
            callback_btn_html=callback_btn_html,
            content="".join(content),
            tooltip_html='<div class="m2g-tooltip" data-toggle="popover" data-content="."></div>'
            if tooltip
            else "",
        )

        # Callback
        if isinstance(callback, _JSCallback):
            if custom_header and callback.library_src:
                custom_header = callback.library_src + custom_header
            else:
                custom_header = callback.library_src
            callback = callback.code
        if callable(callback):
            callback_type = "python"
            cb_handler = partial(callback_handler, callback)
            self.widget.observe(cb_handler, names=["callback_kwargs"])
        else:
            callback_type = "js"

        # Sort
        if sort_by and sort_by != "mols2grid-id":
            if sort_by in (subset + tooltip):
                sort_by = f"data-{slugify(sort_by)}"
            else:
                raise ValueError(
                    f"{sort_by!r} is not an available field in " "`subset` or `tooltip`"
                )
        else:
            sort_by = "mols2grid-id"

        # Slugify remaining vars.
        column_map = {k: slugify(v) for k, v in column_map.items()}
        sort_cols = [slugify(c) for c in sort_cols]
        search_cols = [slugify(c) for c in search_cols]
        smiles = slugify(smiles)
        df = df[final_columns].rename(columns=column_map).sort_values(sort_by)

        template = env.get_template("interactive.html")
        template_kwargs = dict(
            tooltip=tooltip,
            tooltip_placement=repr(tooltip_placement),
            n_items_per_page=n_items_per_page,
            selection=selection,
            truncate=truncate,
            sort_by=sort_by,
            use_iframe=use_iframe,
            #
            border=border,
            gap=gap,
            gap_px="-1px -1px 0 0" if gap == 0 else f"{gap}px",
            pad=pad,
            fontsize=fontsize,
            fontfamily=fontfamily,
            textalign=textalign,
            background_color=background_color,
            hover_color=hover_color,
            #
            iframe_padding=18,
            cell_width=self.img_size[0],
            image_width=self.img_size[0],
            image_height=self.img_size[1],
            #
            item=item,
            item_repr=repr(item),
            value_names=value_names,
            search_cols=search_cols,
            data=json.dumps(
                df.to_dict("records"), indent=None, default=lambda x: "🤷‍♂️"
            ),
            cached_selection=(
                [
                    list(self._cached_selection.keys()),
                    list(self._cached_selection.values()),
                ]
                if self._cached_selection
                else False
            ),
            smiles_col=smiles,
            sort_cols=sort_cols,
            grid_id=self._grid_id,
            whole_cell_style=whole_cell_style,
            custom_css=custom_css or "",
            custom_header=custom_header or "",
            callback=callback,
            callback_type=callback_type,
            removeHs=self.removeHs,
            prefer_coordGen=self.prefer_coordGen,
            onthefly=not self.prerender,
            substruct_highlight=substruct_highlight,
            json_draw_opts=getattr(self, "json_draw_opts", ""),
            single_highlight=single_highlight,
        )
        return template.render(**template_kwargs)

    def get_selection(self):
        """Retrieve the dataframe subset corresponding to your selection.

        Returns
        -------
        pandas.DataFrame
        """
        sel = list(register.get_selection(self._grid_id).keys())
        return self.dataframe.loc[self.dataframe["mols2grid-id"].isin(sel)].drop(
            columns=self._extra_columns
        )

    def filter(self, mask):
        """Filters the grid using a mask (boolean array).

        Parameters
        ----------
        mask : list, pandas.Series or numpy.ndarray
            Boolean array: ``True`` when the item should be displayed,
            ``False`` if it should be filtered out.
        """
        if isinstance(mask, (pd.Series, np.ndarray)):
            mask = mask.tolist()
        if is_running_within_streamlit():
            filtering_script = env.get_template("js/filter.js").render(
                grid_id=self._grid_id, mask=json.dumps(mask)
            )
            return Javascript(filtering_script)
        else:
            self.widget.filter_mask = mask

    def filter_by_index(self, indices):
        """Filters the grid using the dataframe's index."""

        # convert index to mask
        mask = self.dataframe.index.isin(indices)
        return self.filter(mask)

    def to_static(
        self,
        # Display
        subset=None,
        tooltip=None,
        tooltip_fmt="<strong>{key}</strong>: {value}",
        tooltip_trigger="focus",
        tooltip_placement="auto",
        transform=None,
        sort_by=None,
        use_iframe=False,
        truncate=False,
        n_cols=5,
        # CSS Styling
        border="1px solid #cccccc",
        gap=0,
        pad=10,
        fontsize="12px",
        fontfamily="'DejaVu', sans-serif",
        textalign="center",
        custom_css=None,
        style=None,
        # Customization
        custom_header=None,
        **kwargs,
    ):
        """Returns the HTML document for the "static" template

        Parameters: Display
        -------------------
        subset: list or None, default=None
            Columns to be displayed in each cell of the grid. Each column's
            value will be displayed from top to bottom in the order provided.
            The ``"img"`` and ``"mols2grid-id"`` columns are displayed by default,
            however you can still add the ``"img"`` column if you wish to change
            the display order.
        tooltip : list, None or False, default=None
            Columns to be displayed inside the tooltip. When no subset is set,
            all columns will be listed in the tooltip by default. Use ``False``
            to hide the tooltip.
        tooltip_fmt : str, default="<strong>{key}</strong>: {value}"
            Format string of each key/value pair in the tooltip.
        tooltip_trigger : str, default="focus"
            Sequence of triggers for the tooltip: ``click``, ``hover`` or ``focus``
        tooltip_placement : str, default="auto"
            Position of the tooltip: ``auto``, ``top``, ``bottom``, ``left`` or
            ``right``.
        transform : dict or None, default=None
            Functions applied to specific items in all cells. The dict must follow
            a ``key: function`` structure where the key must correspond to one of
            the columns in ``subset`` or ``tooltip``. The function takes the item's
            value as input and transforms it, for example::

                transform={
                    "Solubility": lambda x: f"{x:.2f}",
                    "Melting point": lambda x: f"MP: {5/9*(x-32):.1f}°C"
                }

            These transformations only affect columns in ``subset`` and
            ``tooltip``, and do not interfere with ``style``.
        sort_by : str or None, default=None
            Sort the grid according to the following field (which must be
            present in ``subset`` or ``tooltip``).
        use_iframe : bool, default=False
            Whether to use an iframe to display the grid. When the grid is displayed
            inside a Jupyter Notebook or JupyterLab, this will default to ``True``.
        truncate: bool, default=False
            Whether to truncate the text in each cell if it's too long.
        n_cols : int, default=5
            Number of columns in the table.

        Parameters: CSS
        --------------
        border : str, default="1px solid #cccccc"
            Styling of the border around each cell.
        gap : int, default=0
            Size in pixels of the gap between cells.
        pad: int, default=10
            Size in pixels of the cell padding.
        fontsize : str, default="12pt"
            Font size of the text displayed in each cell.
        fontfamily : str, default="'DejaVu', sans-serif"
            Font used for the text in each cell.
        textalign : str, default="center"
            Alignment of the text in each cell.
        custom_css : str or None, default=None
            Custom CSS properties applied to the generated HTML. Please note that
            the CSS will apply to the entire page if no iframe is used (see
            ``use_iframe`` for more details).
        style : dict or None, default=None
            CSS styling applied to each item in a cell. The dict must follow a
            ``key: function`` structure where the key must correspond to one of the
            columns in ``subset`` or ``tooltip``. The function takes the item's
            value as input, and outputs a valid CSS styling. For example, if you
            want to color the text corresponding to the "Solubility" column in your
            dataframe::

                style={"Solubility": lambda x: "color: red" if x < -5 else ""}

            You can also style a whole cell using the ``__all__`` key, the
            corresponding function then has access to all values for each cell::

                style={"__all__": lambda x: "color: red" if x["Solubility"] < -5 else ""}

        Parameters: Customization
        -----------------------
        custom_header : str or None
            Custom libraries to be loaded in the header of the document

        Returns
        -------
        html_document : str

        Notes
        -----
        If ``subset=None, tooltip=None``, the index and image will be directly
        displayed on the grid while the remaining fields will be in the
        tooltip.

        .. versionadded:: 0.1.0
            Added the ability to style an entire cell with
            ``style={"__all__": <function>}``

        .. versionadded:: 0.2.2
            Added ``sort_by``, ``custom_css``, ``custom_header`` arguments.

        .. versionchanged:: 0.2.2
            If both ``subset`` and ``tooltip`` are ``None``, the index and
            image will be directly displayed on the grid while the remaining
            fields will be in the tooltip.
        """
        if not self.prerender:
            raise ValueError(
                "Please set `prerender=True` when using the 'static' template"
            )
        tr = []
        data = []
        sort_by = sort_by or "mols2grid-id"
        df = self.dataframe.sort_values(sort_by).reset_index(drop=True)

        if subset is None:
            if tooltip is None:
                subset = ["mols2grid-id", "img"]
                tooltip = [x for x in df.columns.tolist() if x not in subset]
            else:
                # When no subset is defined, all columns are displayed.
                subset = df.columns.tolist()
        else:
            # work on a copy
            subset = subset[:]

        if "mols2grid-id" not in subset:
            subset.insert(0, "mols2grid-id")
        if "img" not in subset:
            subset.insert(0, "img")

        # Always make surer the image comes first.
        subset = [subset.pop(subset.index("img"))] + subset

        if style is None:
            style = {}
        if transform is None:
            transform = {}
        if not tooltip:
            tooltip = []

        for i, row in df.iterrows():
            ncell = i + 1
            nrow, ncol = divmod(i, n_cols)
            popover = tooltip_formatter(row, tooltip, tooltip_fmt, style, transform)
            td = [
                f'<td class="col-{ncol} m2g-tooltip" tabindex="0" data-toggle="popover" data-content="{escape(popover)}">'
            ]
            if "__all__" in style.keys():
                s = style["__all__"](row)
                div = [f'<div class="m2g-cell-{i}" style="{s}">']
            else:
                div = [f'<div class="m2g-cell-{i}">']
            for col in subset:
                v = row[col]
                if col == "img" and tooltip:
                    item = f'<div class="data data-img">' f"{v}</div>"
                else:
                    func = style.get(col)
                    slug_col = slugify(col)
                    if func:
                        item = f'<div class="data copy-me data-{slug_col}" style="{func(v)}">'
                    else:
                        item = f'<div class="data copy-me data-{slug_col}">'
                    func = transform.get(col)
                    v = func(v) if func else v
                    item += f"{v}</div>"
                div.append(item)
            div.append("</div>")
            td.append("\n".join(div))
            td.append("</td>")
            tr.append("\n".join(td))

            if (ncell % n_cols == 0) or (ncell == len(df)):
                cell = [f'<tr class="row-{nrow}">']
                cell.append("\n".join(tr))
                cell.append("</tr>")
                data.append("\n".join(cell))
                tr = []

        template = env.get_template("static.html")
        template_kwargs = dict(
            tooltip=tooltip,
            tooltip_trigger=repr(tooltip_trigger),
            tooltip_placement=repr(tooltip_placement),
            use_iframe=use_iframe,
            truncate=truncate,
            #
            border=border,
            gap=gap,
            pad=pad,
            textalign=textalign,
            fontsize=fontsize,
            fontfamily=fontfamily,
            #
            iframe_padding=18,
            cell_width=self.img_size[0],
            #
            custom_css=custom_css or "",
            custom_header=custom_header or "",
            data="\n".join(data),
        )
        return template.render(**template_kwargs)

    @requires("IPython.display")
    def display(
        self,
        use_iframe=False,
        iframe_width="100%",
        iframe_height=None,
        iframe_allow="clipboard-write",
        iframe_sandbox="allow-scripts allow-same-origin allow-downloads allow-popups allow-modals",
        **kwargs,
    ):
        """Render and display the grid in a Jupyter notebook.

        Returns
        -------
        view : IPython.core.display.HTML
        """
        use_iframe = is_jupyter or use_iframe
        doc = self.render(**kwargs, use_iframe=use_iframe)
        if use_iframe:
            # Render HTML in iframe.
            iframe = env.get_template("html/iframe.html").render(
                width=iframe_width,
                height=iframe_height,
                allow=iframe_allow,
                sandbox=iframe_sandbox,
                doc=escape(doc),
            )
            return HTML(iframe)
        else:
            # Render HTML regularly.
            return HTML(doc)

    def save(self, output, **kwargs):
        """Render and save the grid in an HTML document."""
        with open(output, "w", encoding="utf-8") as f:
            f.write(self.render(**kwargs))



import inspect
from functools import singledispatch
from pathlib import Path

from pandas import DataFrame, Series

from .molgrid import MolGrid

_SIGNATURE = {
    method: dict(inspect.signature(getattr(MolGrid, method)).parameters.items())
    for method in ["render", "to_interactive", "to_static", "display"]
}
for method in ["render", "to_interactive", "to_static", "display"]:
    _SIGNATURE[method].pop("self")
    if method in ["render", "display"]:
        _SIGNATURE[method].pop("kwargs")


def _prepare_kwargs(kwargs, kind):
    """Separate kwargs for the init and render methods of MolGrid"""
    template = kwargs.pop("template", _SIGNATURE["render"]["template"].default)
    render_kwargs = {
        param: kwargs.pop(param, sig.default)
        for param, sig in _SIGNATURE[f"to_{template}"].items()
    }
    if kind == "display":
        render_kwargs.update(
            {
                param: kwargs.pop(param, sig.default)
                for param, sig in _SIGNATURE["display"].items()
            }
        )
    return template, kwargs, render_kwargs


@singledispatch
def display(arg, **kwargs):
    """Display molecules on an interactive grid.

    Parameters: Data
    ----------------
    arg : pandas.DataFrame, SDF file or list of molecules
        The input containing your molecules.
    smiles_col : str or None, default="SMILES"
        If a pandas dataframe is used, name of the column with SMILES.
    mol_col : str or None, default=None
        If a pandas dataframe is used, name of the column with RDKit molecules.
        If available, coordinates and atom/bonds annotations from this will be
        used for depiction.

    Parameters: Display
    -------------------
    template : str, default="interactive"
        Either ``"interactive"`` or ``"static"``. See ``render()`` for more details.
    size : tuple, default=(130, 90)
        The size of the drawing canvas. The cell minimum width is set to the
        width of the image, so if the cell padding is increased, the image will
        be displayed smaller.
    useSVG : bool, default=True
        Use SVG images instead of PNG.
    prerender : bool, default=False
        Prerender images for the entire dataset, or generate them on-the-fly.
        Prerendering is slow and memory-hungry, but required when ``template="static"``
        or ``useSVG=False``.
    subset: list or None, default=None
        Columns to be displayed in each cell of the grid. Each column's
        value will be displayed from top to bottom in the order provided.
        The ``"img"`` and ``"mols2grid-id"`` columns are displayed by default,
        however you can still add the ``"img"`` column if you wish to change
        the display order.
    tooltip : list, None or False, default=None
        Columns to be displayed inside the tooltip. When no subset is set,
        all columns will be listed in the tooltip by default. Use ``False``
        to hide the tooltip.
    tooltip_fmt : str, default="<strong>{key}</strong>: {value}"
        Format string of each key/value pair in the tooltip.
    tooltip_trigger : str, default="focus"
        Only available for the "static" template.
        Sequence of triggers for the tooltip: ``click``, ``hover`` or ``focus``
    tooltip_placement : str, default="auto"
        Position of the tooltip: ``auto``, ``top``, ``bottom``, ``left`` or
        ``right``
    transform : dict or None, default=None
        Functions applied to specific items in all cells. The dict must follow
        a ``key: function`` structure where the key must correspond to one of
        the columns in ``subset`` or ``tooltip``. The function takes the item's
        value as input and transforms it, for example::

            transform={
                "Solubility": lambda x: f"{x:.2f}",
                "Melting point": lambda x: f"MP: {5/9*(x-32):.1f}°C"
            }

        These transformations only affect columns in ``subset`` and
        ``tooltip``, and do not interfere with ``style``.
    sort_by : str or None, default=None
        Sort the grid according to the following field (which must be
        present in ``subset`` or ``tooltip``).
    truncate: bool, default=True/False
        Whether to truncate the text in each cell if it's too long.
        Defaults to ``True`` for interactive grids, ``False`` for static grid.
    n_items_per_page, default=24
        Only available for the "interactive" template.
        Number of items to display per page. A multiple of 12 is recommended
        for optimal display.
    n_cols : int, default=5
        Only available for the "static" template.
        Number of columns in the table.
    selection : bool, default=True
        Only available for the "interactive" template.
        Enables the selection of molecules and displays a checkbox at the
        top of each cell. In the context of a Jupyter Notebook, this gives
        you access to your selection (index and SMILES) through
        :func:`mols2grid.get_selection()` or :meth:`MolGrid.get_selection()`.
        In all cases, you can export your selection by clicking on the triple-dot menu.
    cache_selection : bool, default=False
        Only available for the "interactive" template.
        Restores the selection from a previous grid with the same name.
    use_iframe : bool, default=False
        Whether to use an iframe to display the grid. When the grid is displayed
        inside a Jupyter Notebook or JupyterLab, this will default to ``True``.
    iframe_width : str, default="100%
        Width of the iframe
    iframe_height : int or None, default=None
        Height of the frame. When set to ``None``, the height is set dynamically
        based on the content.

    Parameters: Mols
    ----------------
    removeHs : bool, default=False
        Remove hydrogen atoms from the drawings.
    use_coords : bool, default=False
        Use the coordinates of the molecules (only relevant when an SDF file, a
        list of molecules or a DataFrame of RDKit molecules were used as input.)
    coordGen : bool, default=True
        Use the CoordGen library instead of the RDKit one to depict the
        molecules in 2D.
    MolDrawOptions : rdkit.Chem.Draw.rdMolDraw2D.MolDrawOptions or None, default=None
        Drawing options. Useful for making highly customized drawings.
    substruct_highlight : bool or None, default=None
        Highlight substructure when using the SMARTS search. Active by default
        when ``prerender=False``.
    single_highlight : bool, default=False
        Highlight only the first match of the substructure query.

    Parameters: CSS
    ---------------
    border : str, default="1px solid #cccccc"
        Styling of the border around each cell.
    gap : int, default=0
        Size in pixels of the gap between cells.
    pad : int, default=10
        Size in pixels of the cell padding.
    fontsize : str, default="12px"
        Font size of the text displayed in each cell.
    fontfamily : str, default="'DejaVu', sans-serif"
        Font used for the text in each cell.
    textalign : str, default="center"
        Alignment of the text in each cell.
    background_color : str, default="white"
        Only available for the "interactive" template.
        Background color of a cell.
    hover_color : str, default="rgba(0,0,0,0.05)"
        Only available for the "interactive" template.
        Background color when hovering a cell
    custom_css : str or None, default=None
        Custom CSS properties applied to the generated HTML. Please note that
        the CSS will apply to the entire page if no iframe is used (see
        ``use_iframe`` for more details).
    style : dict or None, default=None
        CSS styling applied to each item in a cell. The dict must follow a
        ``key: function`` structure where the key must correspond to one of the
        columns in ``subset`` or ``tooltip``. The function takes the item's
        value as input, and outputs a valid CSS styling. For example, if you
        want to color the text corresponding to the "Solubility" column in your
        dataframe::

            style={"Solubility": lambda x: "color: red" if x < -5 else ""}

        You can also style a whole cell using the ``__all__`` key, the
        corresponding function then has access to all values for each cell::

            style={"__all__": lambda x: "color: red" if x["Solubility"] < -5 else ""}

    Parameters: Customization
    -------------------------
    name : str, default="default"
        Name of the grid. Used when retrieving selections from multiple grids
        at the same time
    rename : dict or None, default=None
        Rename the properties in the final document.
    custom_header : str or None, default=None
        Custom libraries to be loaded in the header of the document.
    callback : str, callable or None, default=None
        Only available for the "interactive" template.
        JavaScript or Python callback to be executed when clicking on an image.
        A dictionnary containing the data for the full cell is directly available
        as ``data`` in JS. For Python, the callback function must have ``data``
        as the first argument to the function. All the values in the ``data`` dict
        are parsed as strings, except "mols2grid-id" which is always an integer.
        Note that fields containing spaces in their name will be replaced by
        hyphens, i.e. "mol weight" becomes available as ``data["mol-weight"]``.

    Returns
    -------
    view : IPython.core.display.HTML

    Notes
    -----
    You can also directly use RDKit's :class:`~rdkit.Chem.Draw.rdMolDraw2D.MolDrawOptions`
    parameters as arguments.
    Additionally, ``atomColourPalette`` is available to customize the atom
    palette if you're not prerendering image (``prerender=False``).

    .. versionadded:: 0.1.0
        Added ``sort_by``, ``custom_css``, ``custom_header`` and ``callback``
        arguments. Added the ability to style an entire cell with
        ``style={"__all__": <function>}``.

    .. versionadded:: 0.2.0
        Added ``substruct_highlight`` argument.

    .. versionchanged:: 0.2.2
        If both ``subset`` and ``tooltip`` are ``None``, the index and image
        will be directly displayed on the grid while the remaining fields will
        be in the tooltip.

    .. versionchanged:: 1.0.0
        ``callback`` can now be a *lambda* function. If ``prerender=True``,
        substructure highlighting will be automatically disabled if it wasn't
        explicitely set to ``True`` instead of raising an error.

    """
    raise TypeError(f"No display method registered for type {type(arg)!r}")


@display.register(DataFrame)
@display.register(dict)
def _(df, **kwargs):
    template, kwargs, render_kwargs = _prepare_kwargs(kwargs, "display")
    return MolGrid(df, **kwargs).display(template=template, **render_kwargs)


@display.register(str)
@display.register(Path)
def _(sdf, **kwargs):
    template, kwargs, render_kwargs = _prepare_kwargs(kwargs, "display")
    return MolGrid.from_sdf(sdf, **kwargs).display(template=template, **render_kwargs)


@display.register(Series)
@display.register(list)
@display.register(tuple)
def _(mols, **kwargs):
    template, kwargs, render_kwargs = _prepare_kwargs(kwargs, "display")
    return MolGrid.from_mols(mols, **kwargs).display(template=template, **render_kwargs)


@singledispatch
def save(arg, **kwargs):
    """Generate an interactive grid of molecules and save it.

    Parameters
    ----------
    arg : pandas.DataFrame, SDF file or list of molecules
        The input containing your molecules.
    output : str
        Name and path of the output document.

    Notes
    -----
    See :func:`display` for the full list of arguments.
    """
    raise TypeError(f"No save method registered for type {type(arg)!r}")


@save.register(DataFrame)
def _(df, **kwargs):
    template, kwargs, render_kwargs = _prepare_kwargs(kwargs, "save")
    output = kwargs.pop("output")
    return MolGrid(df, **kwargs).save(output, template=template, **render_kwargs)


@save.register(str)
@save.register(Path)
def _(sdf, **kwargs):
    template, kwargs, render_kwargs = _prepare_kwargs(kwargs, "save")
    output = kwargs.pop("output")
    return MolGrid.from_sdf(sdf, **kwargs).save(
        output, template=template, **render_kwargs
    )


@save.register(Series)
@save.register(list)
@save.register(tuple)
def _(mols, **kwargs):
    template, kwargs, render_kwargs = _prepare_kwargs(kwargs, "save")
    output = kwargs.pop("output")
    return MolGrid.from_mols(mols, **kwargs).save(
        output, template=template, **render_kwargs
    )



import warnings
from ast import literal_eval


class SelectionRegister:
    """Register for grid selections

    Attributes
    ----------
    SELECTIONS : dict
        Stores each grid selection according to their name
    current_selection : str
        Name of the most recently updated grid
    """

    def __init__(self):
        self.SELECTIONS = {}

    def _update_current_grid(self, name):
        self.current_selection = name

    def _init_grid(self, name):
        overwrite = self.SELECTIONS.get(name, False)
        if overwrite:
            warnings.warn(
                f"Overwriting non-empty {name!r} grid selection: {str(overwrite)}"
            )
        self.SELECTIONS[name] = {}
        self._update_current_grid(name)

    def selection_updated(self, name, event):
        self.SELECTIONS[name] = literal_eval(event.new)
        self._update_current_grid(name)

    def get_selection(self, name=None):
        """Returns the selection for a specific MolGrid instance

        Parameters
        ----------
        name : str or None
            Name of the grid to fetch the selection from. If `None`, the most
            recently updated grid is returned
        """
        name = self.current_selection if name is None else name
        return self.SELECTIONS[name]

    def list_grids(self):
        """Returns a list of grid names"""
        return list(self.SELECTIONS.keys())

    def _clear(self):
        """Clears all selections"""
        if hasattr(self, "current_selection"):
            del self.current_selection
        self.SELECTIONS.clear()


register = SelectionRegister()
get_selection = register.get_selection
list_grids = register.list_grids



import pytest

from mols2grid import callbacks


def test_make_popup_callback():
    popup = callbacks.make_popup_callback(
        title="${title}",
        subtitle="${title}",
        svg="<svg><rect width='10' height='10'/></svg>",
        html='<span id="foo">${title}</span>',
        js='var title = "FOOBAR";',
        style="max-width: 42%;",
    )
    assert '<h2>${title}</h2>' in popup
    assert '<p>${title}</p>' in popup
    assert '<span id="foo">${title}</span>' in popup
    assert '// Prerequisite JavaScript code.\n// prettier-ignore\nvar title = "FOOBAR";' in popup
    assert '<div id="m2g-modal" tabindex="-1" style="max-width: 42%;">' in popup


@pytest.mark.parametrize(
    "title, expected",
    [
        ("SMILES", "${data['SMILES']}"),
        (None, None),
    ],
)
def test_title_field(title, expected):
    assert callbacks._get_title_field(title) == expected



from ast import literal_eval
from base64 import b64decode
from io import BytesIO

import imagehash
from cairosvg import svg2png
from PIL import Image
from selenium import webdriver
from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


class selection_available:
    def __init__(self, is_empty=False):
        self.empty = is_empty

    def __call__(self, driver):
        sel = driver.execute_script("return SELECTION.to_dict();")
        sel = literal_eval(sel)
        if sel == {} and self.empty:
            return True
        elif sel != {} and not self.empty:
            return sel
        return False


class FirefoxDriver(webdriver.Firefox):
    def wait_for_img_load(
        self, max_delay=5, selector="#mols2grid .m2g-cell .data-img svg"
    ):
        return WebDriverWait(self, max_delay).until(
            EC.presence_of_element_located((By.CSS_SELECTOR, selector))
        )

    def wait_for_selection(self, is_empty=False, max_delay=3):
        return WebDriverWait(self, max_delay).until(selection_available(is_empty))

    def wait(self, condition, max_delay=5):
        return WebDriverWait(
            self, max_delay, ignored_exceptions=[StaleElementReferenceException]
        ).until(condition)

    def find_by_id(self, element_id, **kwargs):
        condition = EC.presence_of_element_located((By.ID, element_id))
        return self.wait(condition, **kwargs)

    def find_clickable(self, by, selector, **kwargs):
        condition = EC.element_to_be_clickable((by, selector))
        return self.wait(condition, **kwargs)

    def find_by_css_selector(self, css_selector, **kwargs):
        condition = EC.presence_of_element_located((By.CSS_SELECTOR, css_selector))
        return self.wait(condition, **kwargs)

    def find_by_class_name(self, name, **kwargs):
        condition = EC.presence_of_element_located((By.CLASS_NAME, name))
        return self.wait(condition, **kwargs)

    def find_all_by_class_name(self, name, **kwargs):
        condition = EC.presence_of_all_elements_located((By.CLASS_NAME, name))
        return self.wait(condition, **kwargs)

    def get_svg_hash(self, *args, **kwargs):
        im = next(self.get_imgs_from_svgs(*args, **kwargs))
        return imagehash.average_hash(im, hash_size=16)

    def get_imgs_from_svgs(self, selector="#mols2grid .m2g-cell .data-img"):
        condition = EC.presence_of_all_elements_located((By.CSS_SELECTOR, selector))
        svgs = self.wait(condition)
        for svg in svgs:
            im = svg2png(bytestring=(svg.get_attribute("innerHTML")))
            yield Image.open(BytesIO(im))

    def get_png_hash(self, selector="#mols2grid .m2g-cell .data-img *"):
        img = self.find_by_css_selector(selector)
        im = Image.open(BytesIO(b64decode(img.get_attribute("src")[22:])))
        return imagehash.average_hash(im, hash_size=16)

    def substructure_query(self, smarts):
        self.find_clickable(By.CSS_SELECTOR, "#mols2grid .m2g-search-smarts").click()
        self.find_by_css_selector("#mols2grid .m2g-searchbar").send_keys(smarts)
        self.wait_for_img_load()

    def text_search(self, txt):
        self.find_clickable(By.CSS_SELECTOR, "#mols2grid .m2g-search-text").click()
        self.find_by_css_selector("#mols2grid .m2g-searchbar").send_keys(txt)
        self.wait_for_img_load()

    def clear_search(self):
        self.find_by_css_selector("#mols2grid .m2g-searchbar").clear()
        self.find_by_css_selector("#mols2grid .m2g-searchbar").send_keys(Keys.BACKSPACE)
        self.wait_for_img_load()

    def sort_grid(self, field):
        self.find_clickable(By.CSS_SELECTOR, "#mols2grid .m2g-sort").click()
        self.find_clickable(
            By.CSS_SELECTOR, f'#mols2grid .m2g-sort option[value="data-{field}"]'
        ).click()
        self.wait_for_img_load()

    def invert_sort(self):
        self.find_clickable(By.CSS_SELECTOR, "#mols2grid .m2g-sort .m2g-order").click()
        self.wait_for_img_load()

    def grid_action(self, action):
        self.find_clickable(By.CSS_SELECTOR, "#mols2grid .m2g-actions").click()
        self.find_clickable(
            By.CSS_SELECTOR, f'#mols2grid .m2g-actions option[value="{action}"]'
        ).click()

    def get_tooltip_content(self, pause=0, selector=".m2g-cell .m2g-info"):
        (
            ActionChains(self)
            .move_to_element(self.find_by_css_selector(f"#mols2grid {selector}"))
            .pause(pause)
            .perform()
        )
        tooltip = self.find_by_css_selector('div.popover[role="tooltip"]')
        el = tooltip.find_element(By.CLASS_NAME, "popover-body")
        return el.get_attribute("innerHTML")

    def trigger_callback(
        self, selector="#mols2grid .m2g-cell .m2g-callback", pause=0.2
    ):
        self.wait_for_img_load()
        el = self.find_clickable(By.CSS_SELECTOR, selector)
        (ActionChains(self).move_to_element(el).pause(pause).click().perform())

    def click_checkbox(self, is_empty=False):
        self.find_clickable(By.CSS_SELECTOR, ".m2g-cb").click()
        return self.wait_for_selection(is_empty=is_empty)



from types import SimpleNamespace

import pytest
from numpy.testing import assert_equal
from rdkit import Chem
from rdkit.Chem import Draw
from rdkit.Chem.rdDepictor import Compute2DCoords

from mols2grid import MolGrid
from mols2grid.select import register


def get_grid(df, **kwargs):
    kwargs.setdefault("mol_col", "mol")
    return MolGrid(df, **kwargs)


def test_no_input_specified(small_df):
    with pytest.raises(
        ValueError, match="One of `smiles_col` or `mol_col` must be set"
    ):
        get_grid(small_df, smiles_col=None, mol_col=None)


def test_name_not_str(small_df):
    with pytest.raises(
        TypeError, match="`name` must be a string. Currently of type int"
    ):
        get_grid(small_df, name=0)


def test_uses_svg(grid_prerendered):
    assert grid_prerendered.useSVG == True
    data = grid_prerendered.dataframe.loc[0, "img"]
    assert "<svg " in data[:100]


def test_uses_png(small_df):
    grid = get_grid(small_df, useSVG=False, prerender=True)
    assert grid.useSVG == False
    data = grid.dataframe.loc[0, "img"]
    assert data.startswith('<img src="data:image/png;base64')


def test_copy_df(df):
    grid = MolGrid(df, mol_col="mol")
    assert grid.dataframe is not df


def test_records_as_input(smiles_records):
    MolGrid(smiles_records)


def test_rename(small_df):
    grid = get_grid(small_df, rename={"SOL": "Solubility"})
    assert "Solubility" in grid.dataframe.columns


def test_smi_only_makes_mol_temporary(small_df):
    df = small_df.copy()
    df["SMILES"] = df["mol"].apply(Chem.MolToSmiles)
    df = df.drop(columns=["mol"])
    grid = MolGrid(df)
    assert "mol" not in grid._extra_columns
    assert "mol" not in grid.dataframe.columns


def test_keep_mol_col_if_input(grid_prerendered):
    assert "mol" not in grid_prerendered._extra_columns
    assert "mol" in grid_prerendered.dataframe.columns


def test_remove_hydrogens():
    mol = Chem.MolFromSmiles("C")
    mol = Chem.AddHs(mol)
    data = [{"mol": mol, "ID": 0}]
    grid = MolGrid(data, mol_col="mol", removeHs=True, prerender=True)
    new = grid.dataframe.loc[0, "mol"]
    assert new.GetNumAtoms() == 1


def test_remove_coordinates():
    mol = Chem.MolFromSmiles("C")
    mol = Chem.AddHs(mol)
    Compute2DCoords(mol)
    data = [{"mol": mol, "ID": 0}]
    grid = MolGrid(data, mol_col="mol", use_coords=False, prerender=True)
    new = grid.dataframe.loc[0, "mol"]
    with pytest.raises(ValueError, match="Bad Conformer Id"):
        new.GetConformer()


def test_generate_smi_from_mol_col(small_df):
    df = small_df.drop(columns=["SMILES"])
    grid = MolGrid(df, mol_col="mol", prerender=True)
    assert "SMILES" in grid.dataframe.columns


def test_index_when_none_molecules(small_df):
    df = small_df.copy()
    df.loc[1, "mol"] = None
    grid = MolGrid(df, mol_col="mol", prerender=True)
    assert len(grid.dataframe) == len(df) - 1
    assert_equal(grid.dataframe["mols2grid-id"].values, [0, 2, 3, 4])


def test_custom_moldrawoptions(small_df):
    opts = Draw.MolDrawOptions()
    opts.fixedBondLength = 42
    grid = get_grid(small_df, MolDrawOptions=opts, prerender=True)
    assert grid.MolDrawOptions.fixedBondLength == 42


def test_moldrawoptions_as_kwargs(small_df):
    grid = get_grid(small_df, fixedBondLength=42, prerender=True)
    assert grid.MolDrawOptions.fixedBondLength == 42


def test_update_current_grid_on_init(small_df):
    get_grid(small_df, name="foo")
    assert register.current_selection == "foo"
    assert register.get_selection("foo") is register.get_selection()
    assert register.get_selection() == {}


def test_from_mols(mols):
    grid = MolGrid.from_mols(mols)
    assert "mol" in grid.dataframe.columns


def test_from_mols_custom_mol_col(mols):
    grid = MolGrid.from_mols(mols, mol_col="rdmol")
    assert "rdmol" in grid.dataframe.columns
    assert "mol" not in grid.dataframe.columns


@pytest.mark.parametrize(
    ["param", "attr", "value"],
    [
        ("name", "_grid_id", "foobar"),
        ("coordGen", "prefer_coordGen", False),
        ("removeHs", "removeHs", True),
        ("useSVG", "useSVG", False),
        ("use_coords", "use_coords", False),
        ("size", "img_size", (42, 42)),
        ("prerender", "prerender", True),
        ("smiles_col", "smiles_col", "smi"),
        ("mol_col", "mol_col", "rdmol"),
    ],
)
def test_from_mols_kwargs(mols, param, attr, value):
    if param in ["useSVG"]:
        grid = MolGrid.from_mols(mols, prerender=True, **{param: value})
    else:
        grid = MolGrid.from_mols(mols, **{param: value})
    assert getattr(grid, attr) == value


@pytest.mark.parametrize("sdf_source", ["sdf_path", "sdf_file"])
def test_from_sdf(sdf_source, request):
    sdf = request.getfixturevalue(sdf_source)
    grid = MolGrid.from_sdf(sdf)
    assert "mol" in grid.dataframe.columns


def test_from_sdf_custom_mol_col(sdf_path):
    grid = MolGrid.from_sdf(sdf_path, mol_col="rdmol", prerender=True)
    assert "rdmol" in grid.dataframe.columns
    assert "mol" not in grid.dataframe.columns


@pytest.mark.parametrize(
    ["param", "attr", "value"],
    [
        ("name", "_grid_id", "foobar"),
        ("coordGen", "prefer_coordGen", False),
        ("removeHs", "removeHs", True),
        ("useSVG", "useSVG", False),
        ("use_coords", "use_coords", False),
        ("size", "img_size", (42, 42)),
        ("prerender", "prerender", True),
        ("smiles_col", "smiles_col", "smi"),
        ("mol_col", "mol_col", "rdmol"),
    ],
)
def test_from_sdf_kwargs(sdf_path, param, attr, value):
    if param in ["useSVG"]:
        grid = MolGrid.from_sdf(sdf_path, prerender=True, **{param: value})
    else:
        grid = MolGrid.from_sdf(sdf_path, **{param: value})
    assert getattr(grid, attr) == value


def test_template(grid_prerendered):
    grid_prerendered.template = "interactive"
    assert grid_prerendered._template is grid_prerendered.template


def test_template_setter(grid_prerendered):
    with pytest.raises(ValueError, match="template='foo' not supported"):
        grid_prerendered.template = "foo"


def test_substruct_highlight():
    mol = Chem.MolFromSmiles("C")
    mol.__sssAtoms = [0]
    grid = MolGrid.from_mols([mol], prerender=True)
    svg = grid.dataframe.loc[0, "img"]
    assert "<ellipse" in svg


def test_mol_to_img_svg():
    mol = Chem.MolFromSmiles("C")
    grid = MolGrid.from_mols([mol], prerender=True)
    img = grid.mol_to_img(mol)
    assert "<svg " in img[:100]


def test_mol_to_img_png():
    mol = Chem.MolFromSmiles("C")
    grid = MolGrid.from_mols([mol], prerender=True)
    grid.useSVG = False
    grid._MolDraw2D = Draw.MolDraw2DCairo
    img = grid.mol_to_img(mol)
    assert img.startswith('<img src="data:image/png;base64')


def test_get_selection(small_df):
    grid = MolGrid(small_df, mol_col="mol", name="grid")
    other = MolGrid(small_df, mol_col="mol", name="other")
    event = SimpleNamespace(new='{0: ""}')
    register.selection_updated("grid", event)
    assert register.current_selection == "grid"
    assert grid.get_selection().equals(small_df.head(1))
    assert other.get_selection().equals(small_df.head(0))  # empty dataframe
    register._clear()


def test_save(grid, tmp_path):
    output = tmp_path / "output.html"
    grid.save(output)
    assert output.is_file()


def test_render_interactive(grid):
    grid.render(template="interactive")


def test_render_static(grid_prerendered):
    grid_prerendered.render(template="static")


def test_render_wrong_template(grid):
    with pytest.raises(ValueError, match="template='foo' not supported"):
        grid.render(template="foo")


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(),
        dict(subset=["ID"]),
        dict(tooltip=["ID"]),
        dict(gap=5),
        dict(style={"ID": lambda x: "color: red" if x == 1 else ""}),
        dict(transform={"ID": lambda x: f"Id. #{x}"}),
    ],
)
def test_integration_static(grid_prerendered, kwargs):
    grid_prerendered.to_static(**kwargs)


def test_python_callback(grid):
    html = grid.to_interactive(subset=["img"], callback=lambda data: None)
    assert "// Trigger custom python callback" in html
    assert "// No kernel detected for callback" in html


def test_cache_selection(small_df):
    grid = get_grid(small_df, name="cache")
    event = SimpleNamespace(new='{0: "CCO"}')
    register.selection_updated("cache", event)
    grid = get_grid(small_df, name="cache", cache_selection=True)
    assert hasattr(grid, "_cached_selection")
    assert register.get_selection("cache") == grid._cached_selection


def test_cache_no_init(small_df):
    grid = get_grid(small_df, name="new_cache", cache_selection=True)
    assert hasattr(grid, "_cached_selection")
    assert grid._cached_selection == {}
    assert "new_cache" in register.list_grids()


def test_no_cache_selection(small_df):
    grid = get_grid(small_df, name="no_cache", cache_selection=False)
    assert grid._cached_selection == {}
    assert "no_cache" in register.list_grids()


def test_onthefly_render_png_error(df):
    with pytest.raises(
        ValueError, match="On-the-fly rendering of PNG images not supported"
    ):
        MolGrid(df, prerender=False, useSVG=False)


def test_substruct_highlight_prerender_error(grid_prerendered):
    with pytest.raises(
        ValueError, match="Cannot highlight substructure search with prerendered images"
    ):
        grid_prerendered.display(substruct_highlight=True)
    # disables substruct highlight when prerendering by default
    grid_prerendered.display()


def test_use_coords_onthefly_error(small_df):
    with pytest.raises(
        ValueError, match="Cannot use coordinates with on-the-fly rendering"
    ):
        get_grid(small_df, use_coords=True, prerender=False)


def test_sort_by_not_in_subset_or_tooltip(grid):
    with pytest.raises(ValueError, match="'_Name' is not an available field"):
        grid.to_interactive(subset=["ID", "img"], sort_by="_Name")


def test_subset_without_img_no_error(grid):
    grid.display(subset=["_Name"])


def test_static_no_prerender_error(grid):
    with pytest.raises(
        ValueError, match="Please set `prerender=True` when using the 'static' template"
    ):
        grid.display(template="static")


def test_replace_non_serializable_from_default_output(small_df):
    grid = get_grid(small_df)
    grid.dataframe["non-serializable"] = grid.dataframe[grid.mol_col]
    html = grid.render()
    assert "\\ud83e\\udd37\\u200d\\u2642\\ufe0f" in html  # 🤷‍♂️
